"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
import pytest
from onnx import TensorProto
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info
from onnx.numpy_helper import from_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _build_graph_with_scales(
    mode: str | None = None, ctm: str | None = None, scale: int = 2, ndim: int = 2
):
    attrs: dict = {}
    if mode is not None:
        attrs["mode"] = mode
    if ctm is not None:
        attrs["coordinate_transformation_mode"] = ctm
    resize = make_node(
        "Resize",
        inputs=["x", "", "scales", ""],
        outputs=["y"],
        name="resize",
        **attrs,
    )
    w = 8
    ow = w * scale
    graph = make_graph(
        [resize],
        "graph",
        [make_tensor_value_info("x", TensorProto.FLOAT, (1, 3) + (w,) * ndim)],
        [make_tensor_value_info("y", TensorProto.FLOAT, (1, 3) + (ow,) * ndim)],
        [from_array(np.array([1, 1] + [scale] * ndim, "float32"), name="scales")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return model


def _build_graph_with_sizes(
    mode: str | None = None, ctm: str | None = None, scale: int = 2, ndim: int = 2
):
    attrs: dict = {"axes": list(range(2, 2 + ndim))}
    if mode is not None:
        attrs["mode"] = mode
    if ctm is not None:
        attrs["coordinate_transformation_mode"] = ctm
    resize = make_node(
        "Resize",
        inputs=["x", "", "", "sizes"],
        outputs=["y"],
        name="resize",
        **attrs,
    )
    w = 8
    ow = w * scale
    graph = make_graph(
        [resize],
        "graph",
        [make_tensor_value_info("x", TensorProto.FLOAT, (1, 3) + (w,) * ndim)],
        [make_tensor_value_info("y", TensorProto.FLOAT, (1, 3) + (ow,) * ndim)],
        [from_array(np.array([ow] * ndim, "int64"), name="sizes")],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return model


@pytest.fixture(
    params=[
        # np.random.uniform(0, 255, size=(1, 3, 8)).astype(np.float32),
        np.random.uniform(0, 255, size=(1, 3, 8, 8)).astype(np.float32),
        # np.random.uniform(0, 255, size=(1, 3, 8, 8, 8)).astype(np.float32),
    ]
)
def input_tensors(request):
    """onnxruntime doesn't support interpolate other than 4D tensor"""
    return request.param


@pytest.fixture(params=[2, 3, 4])
def scale_factor(request):
    """Generate scale factors."""
    return request.param


def test_resize_to_convtranspose_nearest_with_scales(input_tensors, scale_factor):
    graph = OnnxGraph(_build_graph_with_scales(mode="nearest", scale=scale_factor))
    runner1 = Evaluator(graph.model)

    pm = PassManager(["resize_to_convtransposed"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    node = graph.nodes["resize"]["pb"]
    assert node.op_type == "ConvTranspose"

    x = input_tensors.copy()
    y = runner1(["y"], {"x": x})[0]
    y_ = runner2(["y"], {"x": x})[0]
    np.testing.assert_allclose(y, y_, rtol=1e-5, atol=1e-5)


def test_resize_to_convtranspose_nearest_with_sizes(input_tensors, scale_factor):
    graph = OnnxGraph(_build_graph_with_sizes(mode="nearest", scale=scale_factor))
    runner1 = Evaluator(graph.model)

    pm = PassManager(["resize_to_convtransposed"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    node = graph.nodes["resize"]["pb"]
    assert node.op_type == "ConvTranspose"

    x = input_tensors
    y = runner1(["y"], {"x": x})[0]
    y_ = runner2(["y"], {"x": x})[0]
    np.testing.assert_allclose(y, y_, rtol=1e-5, atol=1e-5)


@pytest.fixture(
    params=[
        "half_pixel",
        "half_pixel_symmetric",
        "asymmetric",
    ]
)
def ctm_params(request):
    """Generate supported coordinate transformation modes."""
    return request.param


def test_resize_to_convtranspose_linear(input_tensors, ctm_params, scale_factor):
    graph = OnnxGraph(
        _build_graph_with_scales(mode="linear", ctm=ctm_params, scale=scale_factor)
    )
    runner1 = Evaluator(graph.model)

    pm = PassManager(["resize_to_convtransposed"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    node = graph.nodes["resize"]["pb"]
    assert node.op_type == "ConvTranspose"

    # Expect bilinear kernel and pads computed for stride 2
    x = input_tensors
    y = runner1(["y"], {"x": x})[0]
    y_ = runner2(["y"], {"x": x})[0]
    assert y.shape == y_.shape

    slices = [slice(scale_factor - 1, -scale_factor + 1)] * (input_tensors.ndim - 2)
    # psnr
    psnr = 10 * np.log10(255**2 / np.mean((y[..., *slices] - y_[..., *slices]) ** 2))
    assert psnr >= 30, f"PSNR: {psnr:.2f} dB"



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

import numpy as np
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.numpy_helper import from_array, to_array

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _build_concat_conv_model():
    weights = np.arange(4 * 8, dtype=np.float32).reshape(4, 8, 1, 1)
    concat = make_node(
        "Concat",
        inputs=["a", "b"],
        outputs=["concat_out"],
        name="Concat",
        axis=1,
    )
    conv = make_node(
        "Conv",
        inputs=["concat_out", "w"],
        outputs=["y"],
        name="Conv",
        kernel_shape=[1, 1],
    )
    graph = make_graph(
        [concat, conv],
        "pad_concat_inputs",
        [
            make_value_info("a", make_tensor_type_proto(1, [1, 3, 4, 4])),
            make_value_info("b", make_tensor_type_proto(1, [1, 5, 4, 4])),
        ],
        [make_value_info("y", make_tensor_type_proto(1, [1, 4, 4, 4]))],
        [from_array(weights, "w")],
    )
    model = make_model(
        graph,
        ir_version=ONNXIFIER_IR_VERSION,
        opset_imports=[ONNXIFIER_OPSET],
    )
    return model, weights


def test_pad_concat_inputs_rewrites_conv_weights():
    model, weights = _build_concat_conv_model()
    runner_ref = Evaluator(model, "onnx")
    inputs = {
        "a": np.random.uniform(-1, 1, size=[1, 3, 4, 4]).astype(np.float32),
        "b": np.random.uniform(-1, 1, size=[1, 5, 4, 4]).astype(np.float32),
    }
    ref_output = runner_ref(["y"], inputs)[0]

    graph = OnnxGraph(model)
    pm = PassManager(["pad_concat_inputs"])
    optimized = pm.optimize(graph, strict=True)

    concat_node = optimized.nodes["Concat"]["pb"]
    assert list(concat_node.input) == [
        "Concat/pad_a_output0",
        "Concat/pad_b_output0",
    ]

    pad_nodes = [
        optimized.nodes["Concat/pad_a"]["pb"],
        optimized.nodes["Concat/pad_b"]["pb"],
    ]
    assert all(node.op_type == "Pad" for node in pad_nodes)

    conv_node = optimized.nodes["Conv"]["pb"]
    assert conv_node.input[1] == "Conv/padded_weight_output_0"

    weight_constant = optimized.nodes["Conv/padded_weight"]["pb"]
    new_weights = to_array(weight_constant.attribute[0].t)
    assert new_weights.shape == (4, 32, 1, 1)

    expected = np.zeros((4, 32, 1, 1), dtype=weights.dtype)
    expected[:, 0:3, ...] = weights[:, 0:3, ...]
    expected[:, 16:21, ...] = weights[:, 3:8, ...]
    np.testing.assert_allclose(new_weights, expected)

    runner_opt = Evaluator(optimized.model, "onnx")
    opt_output = runner_opt(["y"], {k: v.copy() for k, v in inputs.items()})[0]
    np.testing.assert_allclose(opt_output, ref_output, rtol=1e-5)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

import itertools
import os
import shutil
from copy import deepcopy
from os import PathLike
from pathlib import Path
from typing import Dict, List, Optional, Tuple
from uuid import uuid4

import networkx as nx
from onnx.onnx_pb import FunctionProto, NodeProto

from ... import OnnxGraph, logger
from ...utils import legalize_name
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import extract_function


@PASSES.register("export_functions", patch=["inline_local_functions"])
class ExportFunctionsRewriter(Rewriter):
    """This pass export all functions to plain onnx model.

    Args:
        save_dir (str | PathLike): Specify a directory to save exported functions.
            Functions will be saved in the topological order.
        optimize (bool): Optimize the exported functions.
        recursive (bool): Export functions recursively.
        namespace (str): Namespace of exported functions.

    Example:
        Model definition:
          Model(func1(func2(x)), func3(x))

        Files:
          - 00_func1.onnx
          - 00_00_func2.onnx (if recursive=True)
          - 01_func3.onnx
    """

    def __init__(
        self,
        save_dir: str | PathLike = ".",
        optimize: bool = True,
        recursive: bool = True,
        namespace: str = "",
    ):
        # With pre-order traversal to match nodes in topological order
        super().__init__(SingleNodePattern().with_domain("*").with_order("pre"))
        self.register_post_hook(self._post_hook)
        self.function_map: Dict[str, Tuple[int, OnnxGraph]] = {}
        self.path = Path(save_dir)
        self.optimize = optimize
        self.recursive = recursive
        self.namespace = namespace.strip("/")  # remove leading slash
        self._filelist: List[Path] = []

    def update_io_name(
        self,
        node: NodeProto,
        func: FunctionProto,
        func_nodes: List[NodeProto],
    ):
        """Update io names of nodes in function"""
        outside_io = {name for name in itertools.chain(node.input, node.output)}
        # change the io name which is the same as outside io
        # init: avoid func io is the same as node io
        io_rename = {name: name for name in itertools.chain(func.input, func.output)}
        for n in func_nodes:
            for node_io in itertools.chain(n.input, n.output):
                if node_io in outside_io and node_io not in io_rename:
                    io_rename[node_io] = f"{node_io}-{uuid4().hex[:16]}"

        # func.output may be the input of other nodes in func
        func_io_rename = {
            inside: outside
            for inside, outside in itertools.zip_longest(
                itertools.chain(func.input, func.output),
                itertools.chain(node.input, node.output),
            )
        }

        # update nodes io name
        for n in func_nodes:
            for j, node_input in enumerate(n.input):
                if node_input in io_rename:
                    node_input = io_rename[node_input]
                    n.input[j] = node_input
                if node_input in func_io_rename:
                    n.input[j] = func_io_rename[node_input]

        for n in func_nodes:
            for j, node_output in enumerate(n.output):
                if node_output in io_rename:
                    node_output = io_rename[node_output]
                    n.output[j] = node_output
                if node_output in func_io_rename:
                    n.output[j] = func_io_rename[node_output]

    def rewrite(
        self,
        graph: OnnxGraph,
        nodes: List[NodeProto],
        path: Optional[str | PathLike] = None,
    ):
        if path is not None:
            self.path = path
        node = nodes[0]
        if node.op_type not in graph.functions:
            raise RuntimeError(f"function {node.op_type} not found in the graph")

        model_name = node.name
        if model_name[0] == ".":
            model_name = model_name[1:]
        # prepend an ordered index to set exported file in the same order
        model_name = f"{len(self.function_map):02d}_{model_name}"
        if self.namespace:
            model_name = f"{self.namespace}_{model_name}"
        model_name = legalize_name(model_name)

        # step 1: find the input and output port that matches the custom node
        func = graph.functions[node.op_type]
        func_nodes = [deepcopy(node) for node in func.node]
        self.update_io_name(node, func, func_nodes)

        # step 2: recover constant name
        for func_node in func_nodes:
            if func_node.op_type == "Constant":
                func_node.name = func_node.output[0]

        # step 3: make a copy of each function node appended with identity names
        for i, n in enumerate(func_nodes):
            if not n.name:
                n.name = f"{uuid4().hex}"

        # step 3: assign attributes
        for attr in node.attribute:
            value = self.get_attribute(node, attr.name)
            for n in func_nodes:
                for attr2 in n.attribute:
                    if attr2.ref_attr_name == attr.name:
                        self.set_attribute(n, attr2.name, value)

        self.function_map[model_name] = len(self.function_map), extract_function(
            graph,
            func_nodes,
            # function name has been updated, no mapping needed.
            [(i, i) for i in node.input],
            [(i, i) for i in node.output],
        )

    def _post_hook(self, graph: OnnxGraph):
        for model_name, (order, subgraph) in self.function_map.items():
            from onnxifier import PassManager  # avoid cyclic import

            # cycle check
            cycles = list(nx.simple_cycles(subgraph))
            for cycle in cycles:
                logger.error(f"cycle: {cycle}")
            if self.optimize:
                pm = PassManager(["fold_constant", "eliminate_dead_nodes"])
                subgraph = pm.optimize(subgraph, strict=False)
            if self.recursive and subgraph.functions:
                rewrite_fn = ExportFunctionsRewriter(
                    self.path,
                    self.optimize,
                    self.recursive,
                    f"{self.namespace}/{order:02d}",
                )
                rewrite_fn(subgraph)
                self._filelist.extend(rewrite_fn.filelist)
            else:
                subgraph.save(f"{self.path}/{model_name}.onnx")
                self._filelist.append(Path(f"{self.path}/{model_name}.onnx"))
        # copy external data
        for data in graph.external_data:
            data_path = Path(self.path) / data.name
            if data_path.exists():
                continue
            try:
                os.symlink(data, data_path)
            except (OSError, PermissionError):
                shutil.copyfile(data, data_path)
        return graph

    @property
    def filelist(self) -> List[Path]:
        """List of extracted function files."""
        return self._filelist



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import itertools

from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ...utils import text_to_integers
from .. import OnnxGraph
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="avg_pool")
class AvgPool(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/pooling/avg-pool-14.html

    https://onnx.ai/onnx/operators/onnx__AveragePool.html
    """

    def _trans_ceil_mode(self, ir_mode):
        if ir_mode == "ceil":
            return 1
        return 0

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        ceil_mode = self._trans_ceil_mode(self.get_attribute(ori_node, "rounding_type"))
        pads_begin = self.get_attribute(ori_node, "pads_begin")
        assert isinstance(pads_begin, str) or pads_begin is None
        pads_begin = text_to_integers(pads_begin)
        pads_end = self.get_attribute(ori_node, "pads_end")
        assert isinstance(pads_end, str) or pads_end is None
        pads_end = text_to_integers(pads_end)
        dilations = self.get_attribute(ori_node, "dilations")
        assert isinstance(dilations, str) or dilations is None
        dilations = text_to_integers(dilations)
        strides = self.get_attribute(ori_node, "strides")
        assert isinstance(strides, str) or strides is None
        strides = text_to_integers(strides)
        kernel = self.get_attribute(ori_node, "kernel")
        assert isinstance(kernel, str) or kernel is None
        kernel = text_to_integers(kernel)
        if auto_pad := self.get_attribute(ori_node, "auto_pad"):
            assert isinstance(auto_pad, str)
            if auto_pad == "explicit":
                auto_pad = "NOTSET"
                assert pads_begin is not None and pads_end is not None
                pads = list(itertools.chain(pads_begin, pads_end))
            else:
                pads = None
        else:
            auto_pad = "NOTSET"
            assert pads_begin is not None and pads_end is not None
            pads = list(itertools.chain(pads_begin, pads_end))
        return make_node(
            "AveragePool",
            inputs=ori_node.input,
            outputs=ori_node.output,
            name=ori_node.name,
            auto_pad=auto_pad.upper(),
            ceil_mode=ceil_mode,
            dilations=dilations,
            pads=pads,
            strides=strides,
            kernel_shape=kernel,
        )



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Optional, Sequence

import networkx as nx
import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register("reorder_transpose_to_downstream")
class ReorderTransposeToDownstreamRewriter(Rewriter):
    """Reorder Transpose operator to place it after the element-wise operations.

    Before:

        Transpose -> Add -> Relu -> Mul

    After:

        Add -> Relu -> Mul -> Transpose
    """

    _binary_ops = {"Add", "Sub", "Mul", "Div", "Pow", "Mod", "Exp", "Log"}
    _unary_ops = {"Relu", "LeakyRelu", "PRelu", "Sigmoid", "Tanh", "Erf"}
    _reaxis_ops = {"Concat", "Slice", "Split", "Softmax", "Tile", "Pad"}
    _axis_change_ops = {"Squeeze", "Unsqueeze"}
    _reduce_ops = {"ReduceSum", "ReduceMean", "ReduceMax", "ReduceMin", "ReduceL2"}

    def __init__(self):
        super().__init__(SingleNodePattern("Transpose"))
        self._supported_ops = (
            self._binary_ops.union(self._unary_ops)
            .union(self._reaxis_ops)
            .union(self._axis_change_ops)
            .union(self._reduce_ops)
        )

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        trans_node = nodes[0]
        allowed_nodes = set()
        for i in nx.descendants(graph, trans_node.name):
            if graph.nodes[i]["pb"].op_type in self._supported_ops:
                allowed_nodes.add(i)
        h = graph.onnx_subgraph(allowed_nodes | {trans_node.name})
        allowed_and_connected = set()
        for i in allowed_nodes:
            if nx.has_path(h, trans_node.name, i):
                allowed_and_connected.add(self.graph.nodes[i]["pb"])
        logger.debug(f"# of nodes to reorder: {len(allowed_and_connected)}")
        logger.trace(f"{allowed_and_connected}")
        if not allowed_and_connected:
            return
        h = graph.onnx_subgraph(allowed_and_connected | {trans_node.name})
        if len(h.output) > 1:
            # can't reorder to diverged outputs
            logger.debug(f"Can't reorder to diverged outputs: {len(h.output)}")
            return

        perm = self.get_attribute(trans_node, "perm")
        assert isinstance(perm, list)
        for node in allowed_and_connected:
            if node.op_type in self._unary_ops:
                continue
            elif node.op_type in self._binary_ops:
                tracing = self._get_path(graph, trans_node.name, node.name)
                if node.input[0] in tracing and node.input[1] in tracing:
                    # both inputs are traced, skip
                    continue
                else:
                    self._rewrite_binary_op(
                        graph,
                        node,
                        node.input[1] if node.input[0] in tracing else node.input[0],
                        perm,  # type: ignore
                    )
            elif node.op_type in self._reaxis_ops:
                self._rewrite_reaxis_op(
                    graph,
                    node,
                    perm,  # type: ignore
                )
            else:
                raise NotImplementedError(f"Unsupported op {node.op_type}")

        # reorder transpose node
        for node in graph.onnx_successors(trans_node):
            for i, name in enumerate(node.input):
                if name in trans_node.output:
                    node.input[i] = trans_node.input[0]
        self -= trans_node
        new_trans_node = make_node(
            "Transpose",
            [h.output[0].name],
            trans_node.output,
            name=f"{trans_node.name}/reorder",
            perm=perm,
        )
        self += new_trans_node
        # the last node in subgraph "h"
        out_node = graph.nodes[h._out_to_node[h.output[0].name]]["pb"]
        for node in graph.onnx_successors(out_node):
            for i, name in enumerate(node.input):
                if name in out_node.output:
                    node.input[i] = new_trans_node.output[0]

    def _get_path(self, graph, src, dst):
        traced_path = []
        for path in nx.all_simple_paths(graph, src, dst):
            curr_output = []
            for i in path:
                node = graph.nodes[i]["pb"]
                for j in node.input:
                    if j in curr_output:
                        traced_path.append(j)
                curr_output.extend(node.output)
        return set(traced_path)

    def _rewrite_binary_op(
        self, graph: OnnxGraph, node: NodeProto, another_input: str, perm: List[int]
    ):
        inv_perm = [perm.index(i) for i in range(len(perm))]
        shape = graph.tensor_shape(another_input)
        if len(shape) <= 1:
            return  # safely broadcastable
        value = self.get_value(another_input)
        if len(shape) < len(perm) and value is None:
            raise RuntimeError(
                f"Unsupported node {node.name}: input {another_input} "
                "can not be broadcasted before transpose"
            )
        input_ind = list(node.input).index(another_input)
        if value is not None:
            # fold the constant
            while value.ndim < len(perm):
                value = value[None]
            value = value.transpose(inv_perm)
            cst_node = make_constant(f"{node.name}/{another_input}/const", value)
            node.input[input_ind] = cst_node.output[0]
            self += cst_node
        else:
            trans_node = make_node(
                "Transpose",
                [another_input],
                [f"{another_input}/transpose_output0"],
                name=f"{node.name}/Transpose",
                perm=inv_perm,
            )
            node.input[input_ind] = trans_node.output[0]
            self += trans_node

    def _rewrite_reaxis_op(self, graph: OnnxGraph, node: NodeProto, perm: List[int]):
        axis = self.get_attribute(node, "axis")
        if axis is not None:
            assert isinstance(axis, int)
            if axis < 0:
                axis += len(perm)
            new_axis = perm.index(axis)
            self.set_attribute(node, "axis", new_axis)
        elif node.op_type in ("Pad", "Slice"):
            axes: Optional[Sequence[int] | np.ndarray] = None
            if len(node.input) > 3:
                axes = self.get_value(node.input[3])
            else:
                node.input.append("")
            if axes is None:
                axes = list(range(len(perm)))
            axes = [i + len(perm) if i < 0 else i for i in axes]
            new_axes = [perm.index(i) for i in axes]
            axes_node = make_constant(f"{node.name}/axes", np.array(new_axes, np.int64))
            node.input[3] = axes_node.output[0]
            self += axes_node
            logger.debug(f"node {node.name} new pad axes: {new_axes}")
        elif node.op_type == "Tile":
            repeats = self.get_value(node.input[1])
            if repeats is None:
                raise RuntimeError(
                    f"Tile {node.name} repeats is dynamic, which is not supported"
                )
            new_repeats = [repeats[i] for i in perm]
            cst_node = make_constant(
                f"{node.name}/repeats", np.array(new_repeats, np.int64)
            )
            node.input[1] = cst_node.output[0]
            self += cst_node
            logger.debug(f"node {node.name} new repeats: {new_repeats}")



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ
from typing import List

import networkx as nx
import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("random_initialize_weight", deps=["initializer_to_constant"])
class RandomInitializeWeightRewriter(Rewriter):
    """Randomly initialize Conv and ConvTranspose weights."""

    def __init__(self):
        super().__init__(SingleNodePattern("Conv") | SingleNodePattern("ConvTranspose"))

    def _get_constant(self, graph: OnnxGraph, node):
        pred = self.get_input_node(node, 1)
        assert pred is not None
        for n in nx.ancestors(graph, pred.name):
            if graph.nodes[n]["pb"].op_type == "Constant":
                yield graph.nodes[n]["pb"]

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        for c in self._get_constant(graph, node):
            value = self.get_value(c)
            assert value is not None
            if value.ndim < 4:
                continue
            value = np.random.uniform(
                low=value.min(), high=value.max(), size=value.shape
            ).astype(value.dtype)
            self.set_attribute(c, "value", value)
            logger.debug(f"initialize {c.name} on {node.name}")



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, Sequence

import networkx as nx

from ... import logger
from ...graph import OnnxGraph
from ...passes import PASSES
from . import OP_CONVERTER


def _expand_deps(deps: Sequence[str]):
    root: nx.DiGraph = nx.DiGraph()
    root.add_nodes_from(deps)
    leaves: List[str] = list(deps).copy()
    while leaves:
        leaf = leaves.pop(0)
        children = PASSES[leaf].__DEPS__
        leaves.extend(children)
        root.add_edges_from([(leaf, child) for child in children])
        try:
            cycles = nx.find_cycle(root, leaf)
        except nx.NetworkXNoCycle:
            continue
        else:
            logger.error(f"Cyclic dependencies found!: {cycles}")
            break
    yield from reversed(list(nx.topological_sort(root)))


def downgrade_op_version(graph: OnnxGraph, op_version: int = 17):
    """Downgrade the op version of all nodes in the graph to the specified version."""

    if op_version not in OP_CONVERTER:
        raise NotImplementedError(
            f"Support to downgrade to {sorted(OP_CONVERTER.keys())} for now, "
            f"got {op_version}."
        )
    if graph.opset_version <= op_version:
        return graph
    version_steps = sorted(OP_CONVERTER.keys(), reverse=True)
    for v1, v0 in zip(version_steps, version_steps[1:]):
        if graph.opset_version > v1 and op_version <= v0:
            graph = downgrade_op_version(graph, v1)
            return downgrade_op_version(graph, op_version)

    logger.debug(f"downgrading opset to version {op_version}")
    node_types = set(node["pb"].op_type for node in graph.nodes.values())
    for node_type in node_types:
        if converter := OP_CONVERTER[op_version].get(node_type):
            try:
                for deps in _expand_deps(converter.__DEPS__):
                    graph = PASSES[deps](graph)
                graph = converter(graph)
            except Exception:  # pylint: disable=broad-except
                logger.error(f"Failed to downgrade {node_type} to {op_version}")
                raise
    graph.opset_version = op_version
    return graph



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import onnx.version
from onnx.version_converter import convert_version

from ... import logger
from ...graph import OnnxGraph


def upgrade_op_version(graph: OnnxGraph, op_version: int = 17):
    """Upgrade the op version of all nodes in the graph to the specified version."""

    if graph.opset_version >= op_version:
        return graph

    logger.debug(f"Upgrading opset version from {graph.opset_version} to {op_version}")
    if graph.functions:
        logger.warning(
            "upgrade onnx version will lose functions in the graph, "
            f"this is a bug of onnx {onnx.version.version}. "
            f"Add -v={graph.opset_version} to avoid this auto upgradation."
        )
    new_model = convert_version(graph.model, op_version)
    return OnnxGraph(new_model, base_dir=graph.external_base)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=missing-function-docstring

from openvino import Model
from openvino.opset1 import parameter
from openvino.opset2 import mvn as mvn_1
from openvino.opset6 import mvn as mvn_6

from . import convert_xml


def test_mvn_to_layernorm_opset1_across_channels():
    p = parameter([1, 32, 8, 8])
    r = mvn_1(p, True, True, 1e-5)
    m = Model([r], [p])

    convert_xml(m)


def test_mvn_to_layernorm_opset1_reduction_axes():
    p = parameter([1, 32, 8, 8])
    r = mvn_1(p, False, True, 1e-5)
    m = Model([r], [p])

    convert_xml(m)


def test_mvn_to_layernorm_opset6():
    p = parameter([1, 32, 8, 8])
    r = mvn_6(p, [2, 3], True, 1e-5, "inside_sqrt")
    m = Model([r], [p])

    convert_xml(m)


def test_mvn_to_layernorm_opset6_neg_axes():
    p = parameter([1, 32])
    r = mvn_6(p, [-1], True, 1e-5, "inside_sqrt")
    m = Model([r], [p])

    convert_xml(m)


def test_mvn_no_normalize_opset6():
    p = parameter([1, 32, 8, 8])
    # NOTE: OpenVINO inference is not correct when axes=[2, 3] or axes=[3]
    # There is a bug in onnx frontend that axes are translated to [1,2,3,4]
    # P.S. Onnxruntime inference is correct.
    # https://github.com/openvinotoolkit/openvino/issues/25169
    # NOTE: Fixed after 2024.3
    r = mvn_6(p, [2, 3], False, 1e-5, "inside_sqrt")
    m = Model([r], [p])

    convert_xml(m)


def test_mvn_to_multi_nodes_opset6():
    p = parameter([1, 32, 8, 8])
    r = mvn_6(p, [1, 2, 3], True, 1e-5, "outside_sqrt")
    m = Model([r], [p])

    convert_xml(m)



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Iterable, List, Optional, Tuple

from onnx import NodeProto

from ..graph import OnnxGraph


def find_inbound_nodes(
    graph: OnnxGraph, node: NodeProto, startpoints: Iterable[str]
) -> Tuple[List[NodeProto], List[NodeProto]]:
    r"""Find a set of predecessors of the given node and all of them have a path
    to the given node while the path doesn't contain any node in the given set
    ``startpoints``.

    Example:

           A(Conv)    B(Conv)
             |           |
           C(Relu)       |
             \           /
              \         /
               D(Concat)
                   |
               E(Sigmoid)
                   |
                F(Conv)

        find_inbound_nodes(g, F, ["Conv"]) == ([C, D, E], [A, B])

    Args:
        graph (OnnxGraph): The ONNX graph.
        node (NodeProto): The node to find predecessors for.
        startpoints (Iterable[str]): A set of node names that are not allowed to be
            in the path to the given node.

    Returns:
        Tuple[List[NodeProto]]: A list of found nodes, and a list of start nodes.
    """

    preds = list(graph.onnx_predecessors(node))
    inbounds = []
    start_nodes = []
    while preds:
        pred = preds.pop(0)
        if pred.op_type in set(startpoints):
            start_nodes.append(pred)
        else:
            inbounds.append(pred)
            preds.extend(graph.onnx_predecessors(pred))
    return inbounds, start_nodes


def find_outbound_nodes(
    graph: OnnxGraph, node: NodeProto, endpoints: Iterable[str]
) -> Tuple[List[NodeProto], List[NodeProto]]:
    r"""Find a set of successors of the given node and all of them have a path
    to the given node while the path doesn't contain any node in the given set
    ``endpoints``.

    Example:

           A(Conv)    B(Conv)
             |           |
           C(Relu)       |
             \           /
              \         /
               D(Concat)
                   |
               E(Sigmoid)
                   |
                F(Conv)

        find_outbound_nodes(g, A, ["Conv"]) == ([C, D, E], [F])

    Args:
        graph (OnnxGraph): The ONNX graph.
        node (NodeProto): The node to find successors for.
        endpoints (Iterable[str]): A set of node names that are not allowed to be
            in the path to the given node.

    Returns:
        Tuple[List[NodeProto]]: A list of found nodes, and a list of end nodes.
    """

    succs = list(graph.onnx_successors(node))
    outbounds = []
    end_nodes = []
    while succs:
        succ = succs.pop(0)
        if succ.op_type in set(endpoints):
            end_nodes.append(succ)
        else:
            outbounds.append(succ)
            succs.extend(graph.onnx_successors(succ))
    return outbounds, end_nodes


def find_sibling_nodes(
    graph: OnnxGraph,
    node: NodeProto,
    siblings: Optional[Iterable[str]] = None,
    parents: Optional[Iterable[str]] = None,
) -> List[NodeProto]:
    r"""Find the sibling nodes of the given node in the graph.

    A sibling node is a node that share the same parent node with the given one.

    Example:

             A (Conv)
            /       \
           /         \
        B (Conv)   C (Conv)

        find_sibling_nodes(g, B, ["Conv"]) == [C]

    Args:
        graph (OnnxGraph): a DAG graph.
        node (NodeProto): the given node to find its sibling nodes.
        siblings (Iterable[str], optional): a filter of types to regard as siblings. If
            not provided, the node with same type as the given node are regarded as
            siblings. Defaults to None.
        parents (Iterable[str], optional): a filter of types to regard as
            parents. If not provided, the first predecessor node is regarded as its
            parent. Defaults to None.

    Returns:
        List[NodeProto]: a list of sibling nodes.
    """

    if parents is None:
        parents = [op.op_type for op in graph.onnx_predecessors(node)]
    if siblings is None:
        siblings = [node.op_type]
    else:
        siblings = list(siblings) + [node.op_type]

    _, parent_nodes = find_inbound_nodes(graph, node, parents)
    sibling_nodes = set()
    for n in parent_nodes:
        _, siblings_under_n = find_outbound_nodes(graph, n, siblings)
        sibling_nodes.update(siblings_under_n)

    return list(sibling_nodes.difference([node]))



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from pathlib import Path
from typing import List, Sequence

import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph, logger
from ...utils import legalize_name
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@PASSES.register("save_as_state_dict")
class SaveAsStateDictRewriter(Rewriter):
    """
    Save model as state dict.

    This pass will save the model as a state dict, which is a dictionary that
    maps each parameter name to its value.
    The state dict will be saved in the specified directory.
    """

    def __init__(
        self,
        interested_nodes: Sequence[str] = ("Conv", "ConvTranspose", "MatMul", "Gemm"),
        save_dir: str = ".",
    ):
        patterns = [SingleNodePattern(i) for i in interested_nodes]
        pattern = patterns[0]
        for p in patterns[1:]:
            pattern |= p
        Path(save_dir).mkdir(parents=True, exist_ok=True)
        self.save_dir = save_dir
        self.states: dict[str, np.ndarray] = {}
        super().__init__(pattern=pattern)
        self.register_post_hook(self.save_states)

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]
        name = node.name.strip("/").replace("/", ".")
        if node.op_type in ("Conv", "ConvTranspose"):
            has_bias = len(node.input) > 2 and node.input[2] != ""
            weight = self.get_value_or_die(node.input[1])
            self.states[name + ".weight"] = weight
            if has_bias:
                bias = self.get_value_or_die(node.input[2])
                self.states[name + ".bias"] = bias
        elif node.op_type in ("MatMul", "Gemm"):
            a = self.get_value(node.input[0])
            b = self.get_value(node.input[1])
            if a is None and b is None:
                return
            elif a is not None and b is None:
                self.states[name + ".weight"] = a
            elif b is not None and a is None:
                self.states[name + ".weight"] = b
            elif a is not None and b is not None:
                logger.warning(
                    f"Both inputs to {node.op_type} node '{name}' are constant. "
                    "Saving the second input as '.weight'."
                )
                self.states[name + ".weight"] = b

    def save_states(self, graph: OnnxGraph) -> OnnxGraph:
        try:
            import torch
        except ImportError:
            logger.warning("Please install torch to save model as state dict.")
            return graph

        states = {k: torch.from_numpy(v) for k, v in self.states.items()}
        model_name = legalize_name(graph.name)
        torch.save(states, Path(self.save_dir) / f"{model_name}.pth")
        return graph



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from pathlib import Path
from typing import List, Optional

import onnx

from .intel import IR_DOMAIN

_LAZY_LOAD = {}


def _make_key(operatorset: onnx.OperatorSetIdProto):
    return (operatorset.domain, operatorset.version)


def _lazy_load_xml_frontend():
    # pylint: disable=import-outside-toplevel
    from .intel.openvino.xml_frontend import (
        openvino_xml_to_onnx_graph as _openvino_xml_to_onnx_graph,
    )

    _LAZY_LOAD[_make_key(IR_DOMAIN)] = _openvino_xml_to_onnx_graph


def openvino_xml_to_onnx_graph(
    xml_path: str | os.PathLike | onnx.ModelProto,
    bin_path: Optional[str | os.PathLike] = None,
):
    """Isolate OpenVINO domain unless a IR xml is detected."""
    key = _make_key(IR_DOMAIN)
    if key in _LAZY_LOAD:
        return _LAZY_LOAD[key](xml_path, bin_path)
    raise ImportError("OpenVINO is not imported yet!")


def detect_domain(
    model: str | os.PathLike | onnx.ModelProto,
) -> List[onnx.OperatorSetIdProto]:
    """Detect a custom domain of a model."""

    if isinstance(model, onnx.ModelProto):
        if len(model.opset_import) == 0:
            return []
        opsets = []
        for opset in model.opset_import:
            if opset.domain not in ("", "ai.onnx", "ai.onnx.ml"):
                opsets.append(opset)
            if opset.domain == IR_DOMAIN.domain:
                _lazy_load_xml_frontend()
        return opsets
    model = Path(model).resolve()
    if model.suffix.lower() == ".xml":
        _lazy_load_xml_frontend()
        return [IR_DOMAIN]
    return []


__all__ = ["openvino_xml_to_onnx_graph", "detect_domain"]



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from importlib import import_module
from pathlib import Path
from typing import Optional, Set, Tuple

from ..logger import warning

TOP = Path(__file__).parent.parent.parent.resolve()


def auto_load(cwd: str | Path, filters: Optional[Set[str]] = None, top: Path = TOP):
    """Search and load python modules in `cwd`.

    Args:
        cwd (str | Path): module directory, glob recurrsively
        filters (Set[str], optional): patterns to filter out modules. Defaults to None.
        top (Path, optional): module top directory. Defaults to TOP.
    """
    models_dir = Path(cwd).resolve()
    package = models_dir.relative_to(top).as_posix().replace("/", ".")

    def _filter(p: Path) -> bool:
        pattern: Tuple[str, ...] = ("__",)
        if filters:
            pattern += tuple(filters)
        return all(i not in p.stem for i in pattern)

    for src in filter(_filter, models_dir.rglob("*.py")):
        module_url = src.relative_to(models_dir).with_suffix("").as_posix()
        try:
            import_module(f".{module_url.replace('/', '.')}", package)
        except ImportError as ex:
            warning(f"failed to load {module_url}: {ex}")



"""
Copyright (C) 2024-2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .. import OnnxGraph, make_constant
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="broadcast")
class Broadcast(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/movement/broadcast-3.html

    Use Reshape and Tile to implement broadcast.
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        mode = self.get_attribute(ori_node, "mode")
        if mode not in ("numpy", "explicit", "bidirectional", ""):
            raise NotImplementedError(f"Unsupported mode: {mode}")

        input_shape = graph.tensor_shape(ori_node.input[0])
        input_rank = len(input_shape)
        target_shape = self.get_value(ori_node.input[1])
        if target_shape is None:
            # to expand
            return make_node(
                "Expand",
                inputs=ori_node.input[:2],
                outputs=ori_node.output,
                name=ori_node.name,
            )
        if mode == "bidirectional":
            # recalculate target shape
            for k, (i, j) in enumerate(
                zip(reversed(input_shape), reversed(target_shape))
            ):
                if i != j and i != 1:
                    assert j == 1
                    target_shape[len(target_shape) - k - 1] = i
        elif mode == "explicit":
            axes_mapping = self.get_value_or_die(ori_node.input[2])
            ascending_axes = np.sort(axes_mapping)
            if not np.all(axes_mapping == ascending_axes):
                raise ValueError(
                    f"Expect axes_mapping to be ascending, but got {axes_mapping}"
                )
            shape = np.ones_like(target_shape, dtype=np.int64)
            shape[axes_mapping] = input_shape
            shape_node = make_constant(f"{ori_node.name}/shape", shape)
            reshape = make_node(
                "Reshape",
                inputs=[ori_node.input[0], shape_node.output[0]],
                outputs=[ori_node.input[0] + "/Reshape"],
                name=f"{ori_node.name}/Reshape",
            )
            ori_node.input[0] = reshape.output[0]
            self += [shape_node, reshape]
            input_shape = shape
            input_rank = len(input_shape)
        if input_rank != len(target_shape):
            # insert Unsqueeze to add missing dimensions
            shape = graph.static_tensor_shape(ori_node.input[0])
            shape = [1] * (len(target_shape) - input_rank) + shape
            shape_node = make_constant(
                f"{ori_node.name}/shape", np.array(shape, dtype=np.int64)
            )
            reshape = make_node(
                "Reshape",
                inputs=[ori_node.input[0], shape_node.output[0]],
                outputs=[ori_node.input[0] + "/Reshape"],
                name=f"{ori_node.name}/Reshape",
            )
            ori_node.input[0] = reshape.output[0]
            self += [shape_node, reshape]
            input_shape = shape
        # translate broadcast shape to Tile's repeats
        if np.any(target_shape % input_shape):
            raise ValueError(
                f"Broadcast shape ({target_shape}) is not compatible with "
                f"input shape ({input_shape})."
            )
        repeats = target_shape // input_shape
        repeats_node = make_constant(
            f"{ori_node.name}/repeats", repeats.astype(np.int64)
        )
        ori_node.input[1] = repeats_node.output[0]
        self += repeats_node

        return make_node(
            "Tile",
            inputs=ori_node.input[:2],
            outputs=ori_node.output,
            name=ori_node.name,
        )



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="replicate_identity_constant")
class ReplicateIdentityConstantRewriter(Rewriter):
    """Replicate constant node that identifies to multiple destinations.

    Before:

        constant -> identity -> fanout1
                           | -> fanout2
                           | -> ...

    After:

        constant_copy0 -> fanout1
        constant_copy1 -> fanout2
        constant_copy2 -> ...
    """

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("Identity"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        identity_node = nodes[0]
        constant_node = self.get_input_node(identity_node, 0)
        if constant_node is None:
            # get value from initializer
            value = self.get_value(identity_node.input[0])
            fanout = 0
        elif constant_node.op_type == "Constant":
            value = self.get_value(constant_node)
            fanout = len(graph.onnx_successors(constant_node))
        else:
            return
        if value is None:
            return  # not a constant

        for i, succ in enumerate(graph.onnx_successors(identity_node)):
            pos = list(succ.input).index(identity_node.output[0])
            # make a copy of constant
            constant_copy_node = make_constant(
                f"{identity_node.name}/constant_copy{i}",
                value,
            )
            succ.input[pos] = constant_copy_node.output[0]
            self += constant_copy_node

        self -= identity_node
        if fanout == 1 and isinstance(constant_node, NodeProto):
            self -= constant_node



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor,
    make_tensor_type_proto,
    make_value_info,
)
from onnx.reference import ReferenceEvaluator

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, PassManager
from onnxifier.graph import OnnxGraph


def _build_graph():
    reduce_max = make_node(
        "ReduceMax", ["x", "max_axes"], ["max1"], "reduce_max", keepdims=0
    )
    topk = make_node(
        "TopK",
        ["max1", "topk_k"],
        ["topk1", "topk2"],
        "topk",
        axis=-1,
        largest=1,
        sorted=1,
    )
    unsqueeze = make_node(
        "Unsqueeze", ["topk2", "unsqueeze_axes"], ["unsqueeze1"], "unsqueeze"
    )
    tile = make_node("Tile", ["unsqueeze1", "repeats"], ["tile1"], "tile")
    gather_elements = make_node("GatherElements", ["x", "tile1"], ["y"], axis=1)

    graph = make_graph(
        [reduce_max, topk, unsqueeze, tile, gather_elements],
        "graph",
        # shape[1] of x can't be larger than 32, it will cause a error about np.choose
        # related link:
        # https://stackoverflow.com/questions/39162235/alternative-for-numpy-choose-that-allows-an-arbitrary-or-at-least-more-than-32-a
        [make_value_info("x", make_tensor_type_proto(1, [1, 20, 4]))],
        [make_value_info("y", make_tensor_type_proto(1, [1, 3, 4]))],
        [
            make_tensor("max_axes", 7, [1], np.array([-1], dtype="int64")),
            make_tensor("topk_k", 7, [1], np.array([3], dtype="int64")),
            make_tensor("unsqueeze_axes", 7, [1], np.array([-1], dtype="int64")),
            make_tensor("repeats", 7, [3], np.array([1, 1, 4], dtype="int64")),
        ],
    )
    return make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )


def test_gatherelements_to_gathernd():
    model = _build_graph()
    runner1 = ReferenceEvaluator(model)
    x = np.random.uniform(0, 1, size=[1, 20, 4]).astype(np.float32)
    y1 = runner1.run(None, {"x": x})[0]

    graph = OnnxGraph(model)
    pm = PassManager(["infer_shape", "gatherelements_to_gathernd", "onnxsim"])
    graph = pm.optimize(graph, strict=True)

    # check output
    runner2 = ReferenceEvaluator(graph.model)
    y2 = runner2.run(None, {"x": x})[0]
    assert np.allclose(y1, y2)

    # check conv number
    gather_elements_number = 0
    gatherND_number = 0
    for name in graph:
        if graph.nodes[name]["pb"].op_type == "GatherElements":
            gather_elements_number += 1
        elif graph.nodes[name]["pb"].op_type == "GatherND":
            gatherND_number += 1
    assert gather_elements_number == 0
    assert gatherND_number == 1



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx import numpy_helper
from onnx.helper import (
    make_graph,
    make_model,
    make_node,
    make_tensor_type_proto,
    make_value_info,
)

from onnxifier import ONNXIFIER_OPSET, PassManager
from onnxifier.graph import OnnxGraph


def _build_graph1():
    relu = make_node("Relu", ["x"], ["y"], "relu")
    graph = make_graph(
        [relu],
        "graph",
        [],
        [make_value_info("y", make_tensor_type_proto(1, [1, 3, 256, 256]))],
        [
            numpy_helper.from_array(
                np.random.randn(1, 3, 256, 256).astype(np.float32), "x"
            )
        ],
    )
    return make_model(graph, opset_imports=[ONNXIFIER_OPSET])


def _build_graph2():
    relu = make_node("Relu", ["x"], ["y"], "relu")
    graph = make_graph(
        [relu],
        "graph",
        [],
        [make_value_info("y", make_tensor_type_proto(2, [1, 3, 256, 256]))],
        [
            numpy_helper.from_array(
                np.random.randint(0, 256, size=(1, 3, 256, 256), dtype="uint8"), "x"
            )
        ],
    )
    return make_model(graph, opset_imports=[ONNXIFIER_OPSET])


def test_append_cast_float():
    model = _build_graph1()
    graph = OnnxGraph(model)
    pm = PassManager(["append_cast"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 2
    assert "y/cast" in graph


def test_append_cast_uint8():
    model = _build_graph2()
    graph = OnnxGraph(model)
    pm = PassManager(["append_cast"])
    graph = pm.optimize(graph, strict=True)
    assert len(graph.nodes) == 2
    assert "y/cast" in graph



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ...utils import text_to_boolean, text_to_integers
from .. import OnnxGraph, make_constant
from . import OP_CONVERT, BaseNodeConversion


@OP_CONVERT.register(name="mvn")
class MVN(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/normalization/mvn-6.html

    https://onnx.ai/onnx/operators/onnx__LayerNormalization.html
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        normalize_variance = self.get_attribute(ori_node, "normalize_variance")
        assert isinstance(normalize_variance, str)
        normalize_variance = text_to_boolean(normalize_variance)
        eps = float(self.get_attribute(ori_node, "eps"))  # type: ignore
        eps_mode = self.get_attribute(ori_node, "eps_mode") or "inside_sqrt"
        assert isinstance(eps_mode, str)
        eps_mode = eps_mode.lower()
        assert eps_mode in ("inside_sqrt", "outside_sqrt")

        if len(ori_node.input) == 1:
            # backward compatibility for mvn-1
            across_channels = self.get_attribute(ori_node, "across_channels")
            assert isinstance(across_channels, str)
            across_channels = text_to_boolean(across_channels)
            reduction_axes = self.get_attribute(ori_node, "reduction_axes")
            if across_channels:
                input_rank = len(graph.tensor_shape(ori_node.input[0]))
                axes = np.arange(input_rank, dtype="int64")
            else:
                assert isinstance(reduction_axes, str)
                axes = np.array(text_to_integers(reduction_axes), "int64")
        else:
            axes = self.get_value_or_die(ori_node.input[1]).astype("int64")
        axes_norm = np.arange(axes.min(), axes.max() + 1).astype("int64")
        if len(axes) <= 1:
            rank = len(graph.tensor_shape(ori_node.input[0]))
            axes_norm = np.arange(axes.min(), rank)
        if not np.all(axes == axes_norm) and len(axes) > 1:
            raise ValueError(f"Unsupported axes: {axes}")
        # TODO: axes_norm is no needed, can be replaced by axes after 2024.3
        axes_node = make_constant(f"{ori_node.name}/axes", axes_norm)
        r_mean = make_node(
            "ReduceMean",
            [ori_node.input[0], axes_node.output[0]],
            [f"{ori_node.name}/ReduceMean_output0"],
            name=f"{ori_node.name}/ReduceMean",
            keepdims=1,
        )

        sub = make_node(
            "Sub",
            [ori_node.input[0], r_mean.output[0]],
            [f"{ori_node.name}/Sub_output0"],
            name=f"{ori_node.name}/Sub",
        )

        if not normalize_variance:
            self += [axes_node, r_mean]
            sub.name = ori_node.name
            sub.output[0] = ori_node.output[0]
            return sub
        if eps_mode == "inside_sqrt" and np.all(axes == axes_norm):
            # NOTE: OpenVINO handles LayerNormalization differently with Onnxruntime.
            # In Onnxruntime, `scale` must match the size of normalized layer size.
            # For example, a tensor [1, 32, 8, 8] with `axis=2`, its layer size is 64,
            # so `scale` can be a shape of [64] or [8, 8] or even [1, 64].
            # While in OpenVINO, `scale` must be broadcastable to the layer size. so
            # `scale` must be a shape of [8, 8] in this case.
            axis = int(axes_norm.flatten()[0])
            scale_s = graph.static_tensor_shape(ori_node.output[0])[axis:]
            scale = make_constant(
                f"{ori_node.name}/scale", np.ones(scale_s, np.float32)  # type: ignore
            )
            self += scale
            return make_node(
                "LayerNormalization",
                inputs=[ori_node.input[0], scale.output[0]],
                outputs=[ori_node.output[0]],
                name=ori_node.name,
                axis=axis,
                epsilon=eps,
            )
        mul = make_node(
            "Mul",
            [sub.output[0], sub.output[0]],
            [f"{ori_node.name}/Mul_output0"],
            name=f"{ori_node.name}/Mul",
        )
        r_mean2 = make_node(
            "ReduceMean",
            [mul.output[0], axes_node.output[0]],
            [f"{ori_node.name}/ReduceMean2_output0"],
            name=f"{ori_node.name}/ReduceMean2",
            keepdims=1,
        )
        sqrt = make_node(
            "Sqrt",
            [r_mean2.output[0]],
            [f"{ori_node.name}/Sqrt_output0"],
            name=f"{ori_node.name}/Sqrt",
        )
        eps_node = make_constant(f"{ori_node.name}/eps", np.array(eps, np.float32))
        add_eps = make_node(
            "Add",
            [sqrt.output[0], eps_node.output[0]],
            [f"{ori_node.name}/Add_output0"],
            name=f"{ori_node.name}/Add",
        )
        self += [axes_node, r_mean, sub, mul, r_mean2, eps_node, add_eps, sqrt]
        return make_node(
            "Div",
            inputs=[sub.output[0], add_eps.output[0]],
            outputs=[ori_node.output[0]],
            name=ori_node.name,
        )



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

import math
from typing import List

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(name="gemm_to_conv", deps=["initializer_to_constant"])
class GEMMToConvRewrite(Rewriter):
    """Convert GEMM op to Conv"""

    def __init__(self):
        super().__init__(SingleNodePattern("Gemm") | SingleNodePattern("MatMul"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        gemm_node = nodes[0]
        self._convert_fc_to_conv(graph, gemm_node)
        self._convert_NDim_matmul_to_conv(graph, gemm_node)

    def _convert_fc_to_conv(self, graph: OnnxGraph, gemm_node: NodeProto):
        node = gemm_node.name

        data_shape = graph.static_tensor_shape(gemm_node.input[0])
        weight_shape = graph.static_tensor_shape(gemm_node.input[1])
        alpha = self.get_attribute(gemm_node, "alpha") or 1.0
        beta = self.get_attribute(gemm_node, "beta") or 1.0
        transA = self.get_attribute(gemm_node, "transA") or 0
        transB = self.get_attribute(gemm_node, "transB") or 0
        if len(data_shape) != 2 or len(weight_shape) != 2 or transA:
            return

        # data reshape
        data_shape_cst = make_constant(
            name=f"{node}/DataShape", value=np.array(data_shape + [1, 1], dtype="int64")
        )
        data_reshape_node = make_node(
            op_type="Reshape",
            inputs=[
                gemm_node.input[0],
                data_shape_cst.output[0],
            ],
            outputs=[f"{node}/Conv_input0"],
            name=f"{node}/DataReshape",
        )

        # weight reshape fold
        weight_node = self.get_input_node_or_die(gemm_node, 1)
        if weight_node.op_type == "Constant":
            # fold const
            weight_value = self.get_value_or_die(weight_node).copy()
            weight_value *= alpha
            if not transB:
                weight_value = weight_value.T
            new_weight_node = make_constant(
                name=f"{node}/Weight",
                value=weight_value[..., None, None],
            )
            self -= weight_node
            weight_port = new_weight_node.output[0]
        elif weight_node.op_type == "DequantizeLinear":
            quant_weight_node = self.get_input_node_or_die(weight_node, 0)
            quant_weight = self.get_value_or_die(quant_weight_node)
            if not transB:
                quant_weight = quant_weight.T
                # must quantize on output channels
                assert self.get_attribute(weight_node, "axis") == 1
                self.set_attribute(weight_node, "axis", 0)
            new_weight_node = make_constant(
                f"{node}/QWeight", quant_weight[..., None, None]
            )
            weight_node.input[0] = new_weight_node.output[0]
            weight_port = weight_node.output[0]
            self -= quant_weight_node
            if alpha != 1:
                # rescale with alpha
                scale_node = self.get_input_node_or_die(weight_node, 1)
                scale_value = self.get_value_or_die(scale_node).copy()
                scale_value *= alpha
                scale_const_node = make_constant(f"{node}/scale", scale_value)
                self -= scale_node
                self += scale_const_node
                weight_node.input[1] = scale_const_node.output[0]
        else:
            return

        conv_inputs = [data_reshape_node.output[0], weight_port]

        # bias
        if len(gemm_node.input) == 3:
            bias_node = self.get_input_node_or_die(gemm_node, 2)
            if bias_node.op_type == "Constant":
                bias_value = self.get_value_or_die(bias_node) * beta
                if bias_value.ndim > 1:
                    bias_value = bias_value.squeeze()
                bias_const_node = make_constant(f"{node}/bias", bias_value)
                self -= bias_node
                self += bias_const_node
                conv_inputs.append(bias_const_node.output[0])
            else:
                conv_inputs.append(gemm_node.input[2])

        # conv
        conv_node = make_node(
            op_type="Conv",
            inputs=conv_inputs,
            outputs=[f"{node}/Conv_out"],
            name=f"{node}/Conv",
        )
        self.set_attribute(conv_node, "dilations", [1, 1])
        self.set_attribute(conv_node, "group", 1)
        self.set_attribute(conv_node, "kernel_shape", [1, 1])
        self.set_attribute(conv_node, "pads", [0, 0, 0, 0])
        self.set_attribute(conv_node, "strides", [1, 1])

        # out reshape
        out_shape = graph.static_tensor_shape(gemm_node.output[0])
        out_shape_cst = make_constant(
            name=f"{node}/OutShape", value=np.array(out_shape, dtype="int64")
        )
        out_reshape_node = make_node(
            op_type="Reshape",
            inputs=[
                conv_node.output[0],
                out_shape_cst.output[0],
            ],
            outputs=gemm_node.output[:],
            name=f"{node}/OutReshape",
        )

        self += [
            data_shape_cst,
            data_reshape_node,
            new_weight_node,
            conv_node,
            out_shape_cst,
            out_reshape_node,
        ]
        self -= gemm_node

    def _split_factor(self, number):
        """Split number into two factors, such that the factor1 * factor2 == number"""
        factor = 1
        for i in range(1, int(math.sqrt(number)) + 1):
            if number % i == 0:
                factor = i
        return factor, number // factor

    def _convert_NDim_matmul_to_conv(self, graph: OnnxGraph, gemm_node: NodeProto):
        """It just convert
            [1, B, M, K]x[1, 1, K, N]
        to
            [1, K, B_M', B_M''] @ [N, K, 1, 1]
        It will convert data and weight of matmul to inp and filter of conv
        """
        node = gemm_node.name

        data_shape = graph.static_tensor_shape(gemm_node.input[0])
        weight_shape = graph.static_tensor_shape(gemm_node.input[1])

        # check
        # support unequal number of dimensions
        if len(data_shape) <= 2 and len(weight_shape) <= 2:
            return
        # if both data and weight are batched, it should be split
        data_batch = np.prod(data_shape[:-1])  # type: ignore
        weight_batch = np.prod(weight_shape[:-2])  # type: ignore
        if data_batch != 1 and weight_batch != 1:
            return

        # swap AxB to B'xA'
        swap_operands = data_batch == 1
        # if A is [B, M, K], treat it as [B * M, 1, K]
        B, M, N, K = (
            data_batch * weight_batch,
            1,
            weight_shape[-1],
            data_shape[-1],
        )

        # prepare data input of conv
        if swap_operands:
            # treat weight as input of conv
            inp_input = gemm_node.input[1]
            # [1, 1, M, K]x[1, B, K, N] => [1, K, B, N]@[M, K, 1, 1]
            # [1, B, K, N] => [1, K, B, N]
            n = len(weight_shape)
            perm = [n - 2] + list(range(n - 2)) + [n - 1]
            f1, f2 = self._split_factor(B * N)
        else:
            # treat data as input of conv
            inp_input = gemm_node.input[0]
            # [1, B, M, K]x[1, 1, K, N] => [1, K, B, M]@[N, K, 1, 1]
            # [1, B, M, K] => [1, K, B, M]
            n = len(data_shape)
            perm = [n - 1] + list(range(n - 2)) + [n - 2]
            f1, f2 = self._split_factor(B * M)
        inp_shape = [1, K, f1, f2]
        inp_permute = make_node(
            "Transpose",
            inputs=[inp_input],
            outputs=[f"{node}_data_permute"],
            name=f"{node}/DataPermute",
            perm=perm,
        )
        inp_reshape_cst = make_constant(
            name=f"{node}/DataReshapeCst", value=np.array(inp_shape, dtype="int64")
        )
        inp_reshape = make_node(
            "Reshape",
            inputs=[inp_permute.output[0], inp_reshape_cst.output[0]],
            outputs=[f"{node}_DataReshape"],
            name=f"{node}/DataReshape",
        )
        self += [inp_permute, inp_reshape_cst, inp_reshape]

        # prepare filter of conv
        if swap_operands:
            # treat data as filter of conv
            filter_input = gemm_node.input[0]
            # [1, 1, M, K]x[1, B, K, N] => [1, K, B, N]@[M, K, 1, 1]
            # [1, 1, M, K] => [M, K, 1, 1]
            filter_reshape_cst = make_constant(
                name=f"{node}/filterShape", value=np.array([M, K, 1, 1], dtype="int64")
            )
            filter_node = make_node(
                "Reshape",
                inputs=[filter_input, filter_reshape_cst.output[0]],
                outputs=[f"{node}_filter_reshape"],
                name=f"{node}/FilterReshape",
            )
            self += filter_reshape_cst
        else:
            # treat weight as filter of conv
            filter_input = gemm_node.input[1]
            # [1, B, M, K]x[1, 1, K, N] => [1, K, B, M]@[N, K, 1, 1]
            # [1, 1, K, N] => [N, K, 1, 1]
            perm = [3, 2, 0, 1]
            if len(weight_shape) != 4:
                filter_reshape_cst = make_constant(
                    name=f"{node}/filterShape",
                    value=np.array([1, 1, K, N], dtype="int64"),
                )
                filter_reshape = make_node(
                    "Reshape",
                    inputs=[filter_input, filter_reshape_cst.output[0]],
                    outputs=[f"{node}_filter_reshape"],
                    name=f"{node}/FilterReshape",
                )
                filter_input = filter_reshape.output[0]
                self += [filter_reshape_cst, filter_reshape]
            filter_node = make_node(
                "Transpose",
                inputs=[filter_input],
                outputs=[f"{node}_filter_permute"],
                name=f"{node}/FilterPermute",
                perm=perm,
            )
        self += filter_node

        conv_inputs = [inp_reshape.output[0], filter_node.output[0]]

        # conv
        conv_node = make_node(
            op_type="Conv",
            inputs=conv_inputs,
            outputs=[f"{node}/Conv_out"],
            name=f"{node}/Conv",
            dilations=[1, 1],
            group=1,
            kernel_shape=[1, 1],
            pads=[0, 0, 0, 0],
            strides=[1, 1],
        )

        # out reshape and permute
        out_shape = graph.static_tensor_shape(gemm_node.output[0])  # [1, B, M, N]
        if swap_operands:
            # output of conv is [1, M, B, N]
            perm = [0, 2, 1, 3]
            temp_out_shape = [1, M, B, N]
        else:
            # output of conv is [1, N, B, M]
            perm = [0, 2, 3, 1]
            temp_out_shape = [1, N, B // data_shape[-2], data_shape[-2]]
        temp_out_shape_cst = make_constant(
            name=f"{node}/TempOutShape", value=np.array(temp_out_shape, dtype="int64")
        )
        temp_out_reshape = make_node(
            op_type="Reshape",
            inputs=[
                conv_node.output[0],
                temp_out_shape_cst.output[0],
            ],
            outputs=[f"{node}/temp_out_reshape"],
            name=f"{node}/TempOutReshape",
        )
        self += [
            conv_node,
            temp_out_shape_cst,
            temp_out_reshape,
        ]
        if len(out_shape) == 4:
            out_permute = make_node(
                "Transpose",
                inputs=[temp_out_reshape.output[0]],
                outputs=gemm_node.output[:],
                name=f"{node}/out_permute",
                perm=perm,
            )
            self += out_permute
        else:
            # reshape to NDim array
            out_permute = make_node(
                "Transpose",
                inputs=[temp_out_reshape.output[0]],
                outputs=[f"{node}_out_permute"],
                name=f"{node}/out_permute",
                perm=perm,
            )
            out_shape_cst = make_constant(
                name=f"{node}/OutShape", value=np.array(out_shape, dtype="int64")
            )
            out_reshape_node = make_node(
                op_type="Reshape",
                inputs=[
                    out_permute.output[0],
                    out_shape_cst.output[0],
                ],
                outputs=gemm_node.output[:],
                name=f"{node}/OutReshape",
            )
            self += [
                out_permute,
                out_shape_cst,
                out_reshape_node,
            ]
        self -= gemm_node



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from contextlib import contextmanager
from os import PathLike
from os import chdir as _chdir
from pathlib import Path
from typing import Optional


def legalize_name(name: str) -> str:
    """Replace illegal characters in name with underscores."""
    name = (
        f"{name}".replace(":", "_")
        .replace("/", "_")
        .replace("?", "_")
        .replace("*", "_")
    )
    return name


def legalize_path_name(name: str | PathLike | Path) -> Path:
    """Replace illegal characters in path name with underscores."""
    parent = Path(name).parent
    name = Path(name).name
    name = legalize_name(name)
    return parent / Path(name)


@contextmanager
def chdir(path: Optional[str | PathLike | Path]):
    """Context manager to change current working directory."""
    if path is None:
        yield
        return
    old_cwd = Path.cwd()
    try:
        _chdir(path)
        yield Path(path).resolve()
    finally:
        _chdir(old_cwd)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Optional, Sequence

import numpy as np
from onnx.helper import tensor_dtype_to_np_dtype

from ... import OnnxGraph
from .. import PASSES
from ..utils import make_constant


@PASSES.register()
def trans_input_to_constant(
    graph: OnnxGraph,
    input_name: str | Sequence[str],
    value: Optional[np.ndarray | Sequence[np.ndarray]] = None,
):
    """Consolidate a input to a fixed value as a constant node.

    Args:
        graph (OnnxGraph): onnx graph
        input_name (str | Sequence[str]): input name(s)
        value (np.ndarray | Sequence[np.ndarray], optional): value(s) to be set.
            If None, an all-one array.
    """

    if isinstance(input_name, str):
        input_name = [input_name]
    values: list[None | np.ndarray] = []
    if value is None:
        values = [None for _ in input_name]
    elif not isinstance(value, Sequence):
        values = [value for _ in input_name]
    else:
        values = list(value)  # type: ignore

    assert len(input_name) == len(values)
    for name, v in zip(input_name, values):
        if name not in graph.inputs:
            raise ValueError(f"{name} is not an input of the model")
        shape, ele_type = graph.tensor_info(name)
        if v is None:
            int_shape = graph.static_tensor_shape(name)
            v = np.ones(int_shape, dtype=tensor_dtype_to_np_dtype(ele_type))
        elif shape is not None:
            if len(shape) != v.ndim:
                int_shape = graph.static_tensor_shape(name)
                v = np.broadcast_to(v, int_shape).astype(
                    tensor_dtype_to_np_dtype(ele_type)
                )
            for x, y in zip(shape, v.shape):
                if isinstance(x, int) and x > 0 and x != y:
                    raise ValueError(
                        f"{name} shape {shape} is not compatible with "
                        f"value shape {v.shape}"
                    )
        node = make_constant(name + "/Const", v)
        node.output[0] = name
        graph.add_onnx_node(node)
    return graph



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy as np
import onnx
import pytest
from onnx import TensorProto
from onnx.helper import make_graph, make_model, make_node, make_tensor_value_info

from onnxifier import ONNXIFIER_IR_VERSION, ONNXIFIER_OPSET, OnnxGraph, PassManager
from onnxifier.evaluator import Evaluator


def _build_avgpool(
    kernel: tuple[int, ...] = (3, 3),
    strides: tuple[int, ...] = (1, 1),
    pads: tuple[int, ...] | None = None,
    auto_pad: str | None = None,
    count_include_pad: int = 1,
    ndim: int = 2,
):
    attrs: dict = {"kernel_shape": list(kernel), "strides": list(strides)}
    if pads is not None:
        attrs["pads"] = list(pads)
    if auto_pad is not None:
        attrs["auto_pad"] = auto_pad
    if count_include_pad is not None:
        attrs["count_include_pad"] = int(count_include_pad)

    node = make_node("AveragePool", inputs=["x"], outputs=["y"], name="avg", **attrs)
    w = 8
    # compute output spatial sizes assuming ceil_mode=0 and auto_pad NOTSET
    if pads is None:
        pads_list = [0] * (2 * ndim)
    else:
        pads_list = list(pads)
    out_spatial = []
    for d in range(ndim):
        in_d = w
        k = kernel[d]
        s = strides[d]
        p0 = pads_list[d]
        p1 = pads_list[d + ndim]
        out_d = (in_d + p0 + p1 - k) // s + 1
        out_spatial.append(int(out_d))
    graph = make_graph(
        [node],
        "g",
        [make_tensor_value_info("x", TensorProto.FLOAT, (1, 3) + (w,) * ndim)],
        [make_tensor_value_info("y", TensorProto.FLOAT, (1, 3) + tuple(out_spatial))],
    )
    model = make_model(
        graph, ir_version=ONNXIFIER_IR_VERSION, opset_imports=[ONNXIFIER_OPSET]
    )
    onnx.checker.check_model(model, True)
    return model


@pytest.mark.parametrize("kernel,strides", [((3, 3), (1, 1)), ((2, 2), (2, 2))])
def test_avgpool_to_conv_basic(kernel, strides):
    model = _build_avgpool(kernel=kernel, strides=strides, count_include_pad=1)
    graph = OnnxGraph(model)
    runner1 = Evaluator(graph.model)

    pm = PassManager(["avgpool_to_conv"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    node = graph.nodes["avg"]["pb"]
    assert node.op_type == "Conv"

    x = np.random.randn(1, 3, 8, 8).astype(np.float32)
    y1 = runner1(["y"], {"x": x})[0]
    y2 = runner2(["y"], {"x": x})[0]
    np.testing.assert_allclose(y1, y2, rtol=1e-5, atol=1e-5)


def test_avgpool_to_conv_with_pads_include():
    # count_include_pad=1 with non-zero pads should be supported
    model = _build_avgpool(
        kernel=(3, 3), strides=(2, 2), pads=(1, 1, 1, 1), count_include_pad=1
    )
    graph = OnnxGraph(model)
    runner1 = Evaluator(graph.model)

    pm = PassManager(["avgpool_to_conv"])
    graph = pm.optimize(graph, strict=True)
    runner2 = Evaluator(graph.model)

    node = graph.nodes["avg"]["pb"]
    assert node.op_type == "Conv"

    x = np.random.randn(1, 3, 8, 8).astype(np.float32)
    y1 = runner1(["y"], {"x": x})[0]
    y2 = runner2(["y"], {"x": x})[0]
    np.testing.assert_allclose(y1, y2, rtol=1e-5, atol=1e-5)


def test_avgpool_to_conv_skip_when_exclude_pad_with_nonzero_pad():
    # count_include_pad=0 with pads!=0 cannot be converted; pass should not fire
    model = _build_avgpool(
        kernel=(3, 3), strides=(1, 1), pads=(1, 1, 1, 1), count_include_pad=0
    )
    graph = OnnxGraph(model)

    pm = PassManager(["avgpool_to_conv"])
    graph = pm.optimize(graph, strict=True)

    node = graph.nodes["avg"]["pb"]
    assert node.op_type == "AveragePool"



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(
    "remove_resize_batch_and_channel_scales",
    deps=["resize_move_size_to_scale"],
)
class RemoveResizeBatchAndChannelRewriter(Rewriter):
    """Remove batch and channel scales in the resize node and add axes
    attribute to spatial sizes."""

    def __init__(self):
        super().__init__(SingleNodePattern("Resize"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        if graph.opset_version < 18:
            return  # axes attribute is added since opset 18
        node = nodes[0]
        input_shape = graph.tensor_shape(node.input[0])
        scales = self.get_input_node_or_die(node, 2)
        scales_value = self.get_value(scales)
        if scales_value is None:
            return
        if len(scales_value) != len(input_shape):
            return
        axes = list(range(len(input_shape)))
        if scales_value[0] == 1:
            axes[0] = None  # type: ignore
        if scales_value[1] == 1:
            axes[1] = None  # type: ignore
        axes = list(filter(lambda x: x is not None, axes))
        self.set_attribute(node, "axes", np.array(axes, dtype=np.int64))
        new_scales = make_constant(scales.name + "/new", scales_value[axes])
        node.input[2] = new_scales.output[0]
        self -= scales
        self += new_scales



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import Dict, List, Optional, Tuple

import numpy as np
from onnx.helper import make_node, make_tensor_type_proto, make_value_info

from ... import OnnxGraph
from .. import PASSES


@PASSES.register(deps=["infer_shape"])
def reorder_to_nhwc(
    graph: OnnxGraph, input_names: Optional[List[str]] = None
) -> OnnxGraph:
    """Reorder input tensor to NHWC format by inserting a dummy transpose op.

    Args:
        graph (OnnxGraph): the graph to rewrite
        input_names (List[str], optional): the names of input tensors to reorder.
            Defaults to None, which means all input tensors will be reordered.

    Returns:
        OnnxGraph: the rewritten graph
    """

    input_feeds: Dict[str, Tuple[List[int | str] | None, int]] = {}
    for input_name in graph.inputs:
        input_shape = graph.tensor_shape(input_name)
        if len(input_shape) == 4:
            # this pass only works for NCHW 4D tensor
            if not input_names or input_name in input_names:
                input_feeds[input_name] = graph.tensor_info(input_name)
    if not input_feeds:
        return graph

    for input_name, (input_shape, dtype) in input_feeds.items():
        assert input_shape is not None
        permute = make_node(
            "Transpose",
            inputs=[f"{input_name}_nhwc"],
            outputs=[input_name],
            name=f"nchw_transpose/{input_name}",
            perm=[0, 3, 1, 2],
        )
        graph.input.append(
            make_value_info(
                f"{input_name}_nhwc",
                make_tensor_type_proto(
                    dtype, np.take(input_shape, [0, 2, 3, 1]).tolist()
                ),
            )
        )
        graph.inputs[f"{input_name}_nhwc"] = len(graph.input) - 1
        graph.add_onnx_node(permute)
    return graph



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from collections import defaultdict
from typing import Dict, List

import networkx as nx
import numpy as np
from onnx import numpy_helper
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import PASSES
from ..pattern import GraphPattern, SingleNodePattern, StartEndPointPattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register(deps=["eliminate_dead_nodes", "initializer_to_constant"])
def yolov5_5d_to_4d(graph: OnnxGraph) -> OnnxGraph:  # noqa: C901
    """Convert YOLOv5 5D subgraph to equivalent 4D subgraph."""
    node_to_add = []
    node_to_remove = []
    nodes_5d: Dict[int, List[int]] = defaultdict(list)
    for node_id in nx.topological_sort(graph):
        node_pb = graph.nodes[node_id]["pb"]
        if node_pb.op_type == "Constant":
            continue
        input_5d = []
        output_5d = []
        preds = graph.onnx_predecessors(node_pb)
        preds = [i for i in preds if i.op_type != "Constant"]
        for i in node_pb.input:
            if not i:
                continue
            ndim = len(graph.tensor_shape(i))
            if ndim == 5 and any(i in p.output for p in preds):
                input_5d.append(i)
        for i in node_pb.output:
            if len(graph.tensor_shape(i)) == 5:
                output_5d.append(i)
        if input_5d and output_5d:
            # both input and output is 5D
            if not nodes_5d:
                nodes_5d[node_id].append(node_id)
            else:
                new_head = True
                for head in nodes_5d:
                    if nx.has_path(graph, head, node_id):
                        nodes_5d[head].append(node_id)
                        new_head = False
                        break
                if new_head:
                    nodes_5d[node_id].append(node_id)
    # yolo v5 check
    if len(nodes_5d) != 3:  # 3 branches
        return graph
    for head in nodes_5d:
        head_node = graph.nodes[head]["pb"]
        # insert 4d reshape
        for i, input_name in enumerate(head_node.input):
            shape = graph.static_tensor_shape(input_name)
            shape[3] *= shape.pop(-1)  # [N, C, D, H, W] -> [N, C, D, H*W]
            shape_cst = make_constant(
                f"{head}/Reshape/Const", np.array(shape, dtype="int64")
            )
            reshape_node = make_node(
                "Reshape",
                [input_name, shape_cst.output[0]],
                [f"{head}/Reshape_Output"],
                name=f"{head}/Reshape",
            )
            head_node.input[i] = f"{head}/Reshape_Output"
            node_to_add.extend([reshape_node, shape_cst])
        for node in nodes_5d[head]:
            node_pb = graph.nodes[node]["pb"]
            if node_pb.op_type == "Transpose":
                for attr in node_pb.attribute:
                    if attr.name == "perm":
                        attr.ints.remove(4)
            elif node_pb.op_type in ("Split", "Concat"):
                for attr in node_pb.attribute:
                    if attr.name == "axis":
                        attr.i = 3
            else:
                for pred in graph.onnx_predecessors(node_pb):
                    # replace 5d constant with 4d
                    if pred.op_type == "Constant":
                        data = numpy_helper.to_array(pred.attribute[0].t)
                        shape = list(data.shape)
                        if data.ndim != 5:
                            continue
                        # [N, C, H, W, D] -> [N, C, -1, D]
                        shape = [*shape[:2], shape[2] * shape[3], shape[4]]
                        new_node = make_constant(f"{pred.name}/4D", data.reshape(shape))
                        for i, j in enumerate(node_pb.input):
                            if j == pred.output[0]:
                                node_pb.input[i] = new_node.output[0]
                        node_to_remove.append(pred)
                        node_to_add.append(new_node)

    for node in node_to_remove:
        graph.remove_onnx_node(node)
    for node in node_to_add:
        graph.add_onnx_node(node)
    return graph


@PASSES.register("plate_5d_to_4d")
class Plate5DTo4DRewriter(Rewriter):
    """Rewrite YOLOv5 (Plate Detection) 5D subgraph to equivalent 4D subgraph."""

    def __init__(self):
        gstart = SingleNodePattern("Reshape").with_name("/model.21/Reshape")
        gstart |= SingleNodePattern("Reshape").with_name("/model.21/Reshape_2")
        gstart |= SingleNodePattern("Reshape").with_name("/model.21/Reshape_4")
        gend = GraphPattern()
        r = SingleNodePattern("Reshape")
        gend.add_edge(SingleNodePattern("Concat"), r)
        gend.add_edge(r, SingleNodePattern("Concat"))
        pattern = StartEndPointPattern(
            start_pattern=gstart,
            end_pattern=gend,
        )
        super().__init__(pattern)

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        reshapes = [i for i in nodes if i.op_type == "Reshape"]
        assert len(reshapes) == 2, f"{len(reshapes)} reshapes found, expected 2"
        if not nx.has_path(graph, reshapes[0].name, reshapes[1].name):
            reshapes = [reshapes[1], reshapes[0]]
        assert nx.has_path(graph, reshapes[0].name, reshapes[1].name)
        shape1 = self.get_value_or_die(reshapes[0].input[1])
        new_shape1 = make_constant(
            f"{reshapes[0].name}/reshape",
            np.array([*shape1[:3], shape1[3] * shape1[4]], np.int64),
        )
        reshapes[0].input[1] = new_shape1.output[0]
        self += new_shape1

        transposes = [i for i in nodes if i.op_type == "Transpose"]
        assert len(transposes) == 1
        self.set_attribute(transposes[0], "perm", [0, 1, 3, 2])

        for node in nodes:
            if node.op_type == "Slice":
                axes = self.get_value_or_die(node.input[3])
                assert axes[0] == 4
                new_axes = make_constant(f"{node.name}/axes", np.array([3], np.int64))
                node.input[3] = new_axes.output[0]
                self += new_axes

        concats = [i for i in nodes if i.op_type == "Concat"]
        assert len(concats) == 3
        for node in concats:
            axis = self.get_attribute(node, "axis")
            if axis == 1:
                continue
            self.set_attribute(node, "axis", 3)

        for node in nodes:
            if node.op_type not in ("Mul", "Add"):
                continue
            operand = self.get_value(node.input[1])
            assert operand is not None
            if operand.ndim == 5:
                shape = operand.shape
                operand_node = make_constant(
                    f"{node.name}/operand", operand.reshape([*shape[:2], -1, shape[-1]])
                )
                node.input[1] = operand_node.output[0]
                self += operand_node



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import inspect
from pathlib import Path
from typing import (
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Protocol,
    Sequence,
    Type,
    TypeVar,
    cast,
)

from tabulate import tabulate

from ..traits import RewriterInterface
from .auto_load import auto_load
from .rewriter import Rewriter


class GraphNode(Protocol):
    """Any node to be registered in the Registry shall follow this protocol."""

    __DEPS__: List[str]
    __PATCHES__: List[str]


T = TypeVar("T", bound=GraphNode)
F = TypeVar("F", bound=Callable)


class FuncInterfaceWrapper(Generic[T]):

    def __init__(
        self,
        func: Callable,
        name: Optional[str],
        deps: Optional[List[str]],
        patches: Optional[List[str]],
    ):
        # pylint: disable=invalid-name
        self.__FUNC = func
        self.__NAME__ = name or func.__name__
        self.__DEPS__ = deps or []
        self.__PATCHES__ = patches or []
        setattr(func, "__NAME__", self.__NAME__)
        setattr(func, "__DEPS__", self.__DEPS__)
        setattr(func, "__PATCHES__", self.__PATCHES__)

    def __call__(self) -> T:
        return cast(T, self.__FUNC)


class Registry(Generic[T]):
    """A simple registry object to hold objects from others

    Samples::

        FOO = Registry("FOO")

        @FOO.register()
        def foo(): ...

        print(FOO)
        # ┌───────────────┐
        # │ Register: FOO │
        # ├───────────────┤
        # │ foo           │
        # └───────────────┘
    """

    def __init__(self, name=None, parent: Optional["Registry[T]"] = None) -> None:
        self._bucks: Dict[str, Type[T] | FuncInterfaceWrapper[T]] = {}
        self._configs: Dict = {}
        self._name = name or "<Registry>"
        self._parent = parent
        if parent is not None:
            self._name = f"{parent.name}.{self.name}"

    @property
    def name(self) -> str:
        """Return the name of the registry."""
        return self._name

    @staticmethod
    def _legal_name(name: str) -> str:
        words = [""]
        for a, b in zip(list(name), list(name.lower())):
            if a != b:
                words.append("")
            words[-1] += b
        return "_".join(words).strip("_")

    def register(
        self,
        name: Optional[str] = None,
        deps: Optional[List[str]] = None,
        patch: Optional[List[str]] = None,
    ):
        """A decorator to register an object.

        Args:
            name (str, optional): The name of the object. If not provided, the name
                of the function of class will be used after transform to lowercase.
            deps (List[str], optional): The dependencies before executing the object.
            patch (List[str], optional): The hook after the object execution.
        """

        def wrapper(func: F) -> F:
            if not callable(func):
                raise TypeError(
                    "the object to be registered must be a function or Rewriter,"
                    f" got {type(func)}"
                )
            if inspect.isfunction(func):
                func_wrap = FuncInterfaceWrapper[T](func, name, deps, patch)
                self._bucks[func_wrap.__NAME__] = func_wrap
                self._configs[func_wrap.__NAME__] = inspect.signature(func)
            else:
                assert isinstance(func, type)
                if not issubclass(func, Rewriter):
                    raise TypeError(
                        f"the registered object {func} must be the subclass "
                        f"of Rewriter, but its mro is {func.__mro__}"
                    )

                # note name is not saved because obj is gc-ed after this function
                func.__NAME__ = name or self._legal_name(func.__name__)
                func.__DEPS__.extend(deps or [])
                func.__PATCHES__.extend(patch or [])
                self._bucks[func.__NAME__] = cast(Type[T], func)
                self._configs[func.__NAME__] = inspect.signature(func.rewrite)
            if self._parent is not None:
                self._parent.register(name, deps, patch)(func)
            # forward the signature of the original function
            return cast(F, func)

        return wrapper

    def get(self, name: str) -> Optional[T]:
        """Get a registered object by its name."""
        if name in self._bucks:
            functor = self._bucks[name]()  # create a new instance each time
            # functor.__NAME__ = name  # rename the instance
            return functor

    def get_config(self, name: str):
        """Get the configuration of an object"""
        return self._configs.get(name)

    def child(self, passes: str | Sequence[str]) -> "Registry":
        """Slice a child registry by given a set of pass names."""

        reg = self.__class__(parent=self)
        if isinstance(passes, str):
            passes = [passes]
        # pylint: disable=protected-access
        reg._bucks = {k: self._bucks[k] for k in passes}
        reg._configs = {k: self._configs[k] for k in passes}
        return reg

    def __getitem__(self, name: str) -> T:
        """Get a registered object by its name."""
        obj = self.get(name)
        if obj is None:
            raise KeyError(f"{name} is not registered in {self._name}")
        return obj

    def __iter__(self) -> Iterator[str]:
        """Return an Iterator for all registered functions"""
        yield from self._bucks.keys()

    def __contains__(self, name: str) -> bool:
        """Check if a function is registered"""
        return name in self._bucks

    def __repr__(self) -> str:
        title = [f"Register: {self._name}", "Deps", "Patch", "Config"]
        members = []
        for i in sorted(self._bucks.keys()):
            members.append(
                [
                    i,
                    self._bucks[i].__DEPS__,
                    self._bucks[i].__PATCHES__,
                    self._configs[i],
                ]
            )
        return tabulate(members, title, "simple_grid", maxcolwidths=[None, 50, 50, 50])


def get_pass_manager(
    include: Optional[Sequence[str]] = None,
    exclude: Optional[Sequence[str]] = None,
    configs: Optional[Dict[str, Dict[str, str | int | float | bool]]] = None,
):
    """Lazy load pass manager"""
    # pylint: disable=import-outside-toplevel
    from ..pass_manager import PassManager

    return PassManager(include, exclude, configs)


PASSES = Registry[RewriterInterface]("PASS")
L1 = Registry[RewriterInterface]("L1", parent=PASSES)
L2 = Registry[RewriterInterface]("L2", parent=PASSES)
L3 = Registry[RewriterInterface]("L3", parent=PASSES)

_AUTO_LOAD_FOLDERS = sorted(filter(Path.is_dir, Path(__file__).parent.glob("*")))
for i in _AUTO_LOAD_FOLDERS:
    auto_load(i)



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

# pylint: disable=arguments-differ

from typing import List

from onnx.onnx_pb import NodeProto

from ... import OnnxGraph
from .. import L2
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter


@L2.register(name="eliminate_identity", deps=["replicate_identity_constant"])
class EliminateIdentityRewriter(Rewriter):
    """Remove Identity nodes."""

    def __init__(self):
        super().__init__(pattern=SingleNodePattern("Identity"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        identity_node = nodes[0]
        if graph.nodes[identity_node.name]["has_output"]:
            prev_node = graph.onnx_predecessors(identity_node)[0]
            for i, output_name in enumerate(prev_node.output):
                if output_name == identity_node.input[0]:
                    prev_node.output[i] = identity_node.output[0]
            # Since identity's input is replaced by its output, every sibling node's
            # input should also be replaced.
            sibling_nodes = graph.onnx_siblings(identity_node)
            for node in sibling_nodes:
                for i, input_name in enumerate(node.input):
                    if input_name == identity_node.input[0]:
                        node.input[i] = identity_node.output[0]
            # replace output node
            graph.set_output(prev_node, identity_node.output[0])
            self -= identity_node
            return
        for node in graph.onnx_successors(identity_node):
            for i, input_name in enumerate(node.input):
                if input_name == identity_node.output[0]:
                    node.input[i] = identity_node.input[0]
        self -= identity_node



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List

import numpy as np
from onnx.helper import make_node
from onnx.onnx_pb import NodeProto

from .... import OnnxGraph
from ...pattern import SingleNodePattern
from ...rewriter import Rewriter
from ...utils import make_constant
from . import OP_CONVERTER


@OP_CONVERTER.register("CenterCropPad")
class CenterCropPad(Rewriter):
    """CenterCropPad to pad and slice"""

    def __init__(self):
        super().__init__(SingleNodePattern("CenterCropPad"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto], *args, **kwargs):
        node = nodes[0]
        axes = self.get_attribute(node, "axes")
        shape = self.get_value(node.input[1])
        if shape is None:
            raise RuntimeError(f"shape of {node.name} is not constant")
        ori_shape = graph.static_tensor_shape(node.input[0])
        if axes is None:
            axes = list(range(len(shape)))
        axes = sorted(axes)  # type: ignore

        if self._need_pad(shape, ori_shape, axes):
            shape_iter = iter(shape)
            shape_expand = list(ori_shape)
            for i in axes:
                shape_expand[i] = next(shape_iter)
            pad_list = [max(0, i - j) for i, j in zip(shape_expand, ori_shape)]
            paddings = np.array([[i // 2, i - i // 2] for i in pad_list], np.int64)
            paddings = paddings.transpose().flatten()
            pad_cst = make_constant(f"{node.name}/Pad/pads", paddings)
            pad = make_node(
                "Pad",
                inputs=[node.input[0], pad_cst.output[0]],
                outputs=[f"{node.name}/Pad_output0"],
                name=f"{node.name}/Pad",
            )
            self += [pad_cst, pad]
            node.input[0] = pad.output[0]
            ori_shape = [max(i, j) for i, j in zip(shape_expand, ori_shape)]
        # get slice begin and end
        begin, end = self._slice_begin_and_end(shape, ori_shape, axes)
        beg_node = make_constant(f"{node.name}/Slice/begin", np.array(begin, np.int64))
        end_node = make_constant(f"{node.name}/Slice/end", np.array(end, np.int64))
        axes_node = make_constant(f"{node.name}/Slice/axes", np.array(axes, np.int64))
        slice_node = make_node(
            "Slice",
            inputs=[
                node.input[0],
                beg_node.output[0],
                end_node.output[0],
                axes_node.output[0],
            ],
            outputs=node.output,
            name=node.name,
        )
        self += [beg_node, end_node, axes_node, slice_node]
        self -= node

    def _need_pad(self, shape, ori_shape, axes):
        shape_iter = iter(shape)
        return any(next(shape_iter) > ori_shape[i] for i in axes)

    def _slice_begin_and_end(self, shape, ori_shape, axes):
        shape_iter = iter(shape)
        slices = [ori_shape[i] - next(shape_iter) for i in axes]
        begin = [s // 2 for s in slices]
        end = [ori_shape[i] - s // 2 for i, s in zip(axes, slices)]
        return begin, end



"""
Copyright (C) 2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest


def classification_models():
    """A list of testing classification models from onnx hub."""
    header = "https://github.com/onnx/models/raw/main/validated/vision/classification/"
    models = [
        ("mobilenet/model/mobilenetv2-10.onnx", None),
        ("resnet/model/resnet18-v2-7.onnx", None),
        ("resnet/model/resnet50-v1-12-qdq.onnx", None),
    ]
    return [(header + i + "?download=", hash_value) for (i, hash_value) in models]


def pytest_generate_tests(metafunc: pytest.Metafunc):
    """Generate parametrized arguments to all tests with arg 'model'.
    `model` is acquired from model_zoo.
    """

    if "classification_models" in metafunc.fixturenames:
        metafunc.parametrize("classification_models", classification_models())



"""
Copyright (C) 2024 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from onnx import numpy_helper
from onnx.helper import make_node, tensor_dtype_to_np_dtype
from onnx.onnx_pb import NodeProto
from openvino import Model, compile_model
from openvino.opset8 import constant, prior_box

from ...utils import text_to_boolean
from .. import OnnxGraph
from . import OP_CONVERT, BaseNodeConversion


def _to_floats(text):
    if not text:
        return []
    return list(map(float, text.split(",")))


@OP_CONVERT.register(name="prior_box")
class PriorBox(BaseNodeConversion):
    """https://docs.openvino.ai/2024/documentation/openvino-ir-format/operation-sets/operation-specs/detection/prior-box-8.html

    Fold to constant
    """

    def replace(self, graph: OnnxGraph, ori_node: NodeProto) -> NodeProto:
        layer_shape = self.get_value(ori_node.input[0])
        image_shape = self.get_value(ori_node.input[1])
        if layer_shape is None or image_shape is None:
            return ori_node
        out_shape, out_dtype = graph.tensor_info(ori_node.output[0])
        layer_shape = constant(layer_shape, name="layer_shape")
        image_shape = constant(image_shape, name="image_shape")

        pb_attrs = dict(
            min_size=_to_floats(self.get_attribute(ori_node, "min_size")),
            max_size=_to_floats(self.get_attribute(ori_node, "max_size")),
            aspect_ratio=_to_floats(self.get_attribute(ori_node, "aspect_ratio")),
            density=_to_floats(self.get_attribute(ori_node, "density")),
            fixed_ratio=_to_floats(self.get_attribute(ori_node, "fixed_ratio")),
            fixed_size=_to_floats(self.get_attribute(ori_node, "fixed_size")),
            clip=text_to_boolean(self.get_attribute(ori_node, "clip")),  # type: ignore
            flip=text_to_boolean(self.get_attribute(ori_node, "flip")),  # type: ignore
            step=float(self.get_attribute(ori_node, "step")),  # type: ignore
            offset=float(self.get_attribute(ori_node, "offset") or 0),  # type: ignore
            variance=_to_floats(self.get_attribute(ori_node, "variance")),
            scale_all_sizes=text_to_boolean(
                self.get_attribute(ori_node, "scale_all_sizes")  # type: ignore
            ),
        )
        if order := self.get_attribute(ori_node, "min_max_aspect_ratios_order"):
            assert isinstance(order, str)
            pb_attrs["min_max_aspect_ratios_order"] = text_to_boolean(order)
        prior_node = prior_box(layer_shape, image_shape, pb_attrs)
        model = Model([prior_node], [])  # type: ignore
        compiled_model = compile_model(model, "CPU")
        value = compiled_model()[-1]
        assert out_shape is not None and value.shape == tuple(out_shape)
        assert value.dtype == tensor_dtype_to_np_dtype(out_dtype)

        return make_node(
            "Constant",
            [],
            outputs=ori_node.output,
            name=ori_node.name,
            value=numpy_helper.from_array(value),
        )



"""
Copyright (C) 2024-2026 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import gc
import os
import traceback
import warnings
from copy import deepcopy
from io import IOBase
from itertools import chain
from pathlib import Path
from typing import IO, Dict, Iterable, List, Optional, Sequence, Tuple
from uuid import uuid4

import networkx as nx
import onnx
from onnx import external_data_helper as edh
from onnx.helper import (
    make_graph,
    make_model,
    make_operatorsetid,
    make_tensor_sequence_value_info,
    make_tensor_type_proto,
    make_tensor_value_info,
    make_value_info,
)

from .logger import debug, error
from .utils import chdir


def _unique_opset(opset_import: Sequence[onnx.OperatorSetIdProto]):
    domain_version: Dict[str, int] = {}
    for opset in opset_import:
        domain_version[opset.domain] = max(
            domain_version.get(opset.domain, 1), opset.version
        )
    return [make_operatorsetid(k, v) for k, v in domain_version.items()]


class OnnxGraph(nx.DiGraph):
    """Create a DAG from onnx graph.

    Args:
        model (onnx.ModelProto): use onnx.load_model to get the proto.
        base_dir (str): If the model contains external data, specify the base directory
            to load external data that relative to. If not specified, the current
            working directory is used as the base directory.
    """

    def __init__(
        self,
        model: Optional[onnx.ModelProto] = None,
        base_dir: Optional[str] = None,
    ) -> None:
        if model is None:
            super().__init__()
        else:
            assert isinstance(model, onnx.ModelProto)
            super().__init__(name=model.graph.name)
            self._model = model
            self._node_to_out: Dict[str, Sequence[str]] = {}
            self._out_to_node: Dict[str, str] = {}
            self._functions = {f.name: f for f in model.functions}
            try:
                graph = onnx.shape_inference.infer_shapes(model).graph
            except onnx.shape_inference.InferenceError as ex:
                warnings.warn(f"Inferring shape failed, value info is not set:\n {ex}")
                graph = model.graph
            # WA: infer_shapes could cause graph to be empty in some cases
            if len(graph.node) == 0:
                graph = model.graph
            self.inputs = {i.name: n for n, i in enumerate(graph.input)}
            self.outputs = {i.name: n for n, i in enumerate(graph.output)}
            # read-only value info
            self._value_info = graph.value_info
            # write-only value info
            self._value_info_update: List[onnx.ValueInfoProto] = []
            self._keep_value_info = False
            for node in graph.node:
                self._add_onnx_node_internal(node)
            self._build_edges()
            # OnnxGraph save external tensors to one file, and this variable keeps
            # the parent directory of the tensor file, which is used to specify
            # a correct working directory when loading external data into onnx model.
            self._external_base: Optional[str] = self._get_external_base(base_dir)

    def _get_external_base(self, base_dir: str | None) -> str | None:
        err_msg = ""
        for t in self._get_all_tensors():
            if edh.uses_external_data(t):
                info = edh.ExternalDataInfo(t)
                if base_dir is None:
                    base_dir = info.basepath
                elif info.basepath and base_dir != info.basepath:
                    raise ValueError("External data have different base path!")
                if not os.path.exists(os.path.join(base_dir, info.location)):
                    err_msg += f"External {t.name} @{info.location} not found!\n"
        if err_msg:
            raise FileNotFoundError(err_msg)
        if base_dir == "":
            # WA: actually `basepath` is not used in current onnx
            base_dir = Path.cwd().as_posix()
        return base_dir

    def _add_onnx_node_internal(self, node: onnx.NodeProto) -> None:
        if not node.HasField("name"):
            node.name = f"{node.op_type}/{uuid4().hex}"
        name = node.name
        has_input = any(i in self.inputs for i in node.input)
        has_output = any(i in self.outputs for i in node.output)
        # clear out-dated edges on this node
        if name in self:
            self.remove_node(name)
        self.add_node(name, pb=node, has_input=has_input, has_output=has_output)
        self._node_to_out[name] = node.output
        for o in node.output:
            # we need handle cases that node A replaces node B, so _out_to_node
            # should change from B to A. Previously this behavior is UB.
            if o in self._out_to_node and self._out_to_node[o] != name:
                debug(
                    f"{self._out_to_node[o]} shares output {o} with {name}, "
                    "make sure the former will be removed."
                )
            self._out_to_node[o] = name

    def _build_edges(self):
        for n in self.nodes:
            node = self.nodes[n]["pb"]
            for i in node.input:
                if upstream := self._out_to_node.get(i):
                    self.add_edge(upstream, n, edge=i)

    def copy(self, as_view=False, onnx_copy=False):
        """Create a deep copy of the OnnxGraph.

        This method overrides NetworkX's copy method to ensure all ONNX-specific
        attributes are properly copied.

        Args:
            as_view (bool): If True, return a view (not implemented for OnnxGraph).
                           Defaults to False.
            onnx_copy (bool): If True, perform a deep copy of the model.
                           Defaults to False.
        """
        if not onnx_copy:
            return super().copy(as_view=as_view)
        if as_view:
            raise NotImplementedError("View copying is not supported for OnnxGraph")

        # Create a deep copy of the model to ensure all ONNX data is preserved
        copied_model = deepcopy(self._model)

        # Create a new OnnxGraph from the copied model
        copied_graph = OnnxGraph(copied_model, base_dir=self._external_base)

        # Copy any additional attributes that might have been modified
        copied_graph._keep_value_info = getattr(self, "_keep_value_info", False)
        copied_graph._value_info_update = deepcopy(self._value_info_update)

        return copied_graph

    def add_onnx_node(self, node: onnx.NodeProto) -> None:
        """Insert a node and its edges."""
        assert isinstance(node, onnx.NodeProto)
        self._add_onnx_node_internal(node)
        self._build_edges()
        # remove connected edge from inputs
        for output_name in node.output:
            for i, graph_input in enumerate(self.input):
                if graph_input.name == output_name:
                    self.input.pop(i)
                    self.inputs.pop(output_name)

    def remove_onnx_node(self, node: onnx.NodeProto | str, no_replace=False) -> None:
        """Remove a node from the graph"""
        assert isinstance(node, (onnx.NodeProto, str))
        if isinstance(node, str):
            name = node
            node = self.nodes[name]["pb"]
        else:
            name = node.name
        if name not in self:
            raise ValueError(f"Node {name} is not existing in the graph!")
        assert isinstance(node, onnx.NodeProto)
        if no_replace:
            # dump node output info to inputs
            for output_name in node.output:
                if self._out_to_node.get(output_name) != name:
                    # node is replaced
                    continue
                shape, dtype = self.tensor_info(output_name)
                self.input.append(
                    make_value_info(output_name, make_tensor_type_proto(dtype, shape))
                )
                self.inputs[output_name] = len(self.input) - 1
        self._node_to_out.pop(name)
        for output_name in node.output:
            if self._out_to_node[output_name] == name:
                self._out_to_node.pop(output_name)
                # output is not claimed by added node, remove it from graph
                if output_name in self.outputs:
                    self.remove_output(output_name)
        self.remove_node(name)

    def onnx_predecessors(self, n: onnx.NodeProto | str) -> List[onnx.NodeProto]:
        """Returns a list of predecessor nodes of n."""
        assert isinstance(n, (onnx.NodeProto, str))
        preds = self.predecessors(n if isinstance(n, str) else n.name)
        return [self.nodes[p]["pb"] for p in preds]

    def onnx_successors(self, n: onnx.NodeProto | str) -> List[onnx.NodeProto]:
        """Returns a list of successors nodes of n."""
        assert isinstance(n, (onnx.NodeProto, str))
        succs = self.successors(n if isinstance(n, str) else n.name)
        return [self.nodes[s]["pb"] for s in succs]

    def onnx_siblings(self, n: onnx.NodeProto | str) -> List[onnx.NodeProto]:
        r"""Returns a list of sibling nodes of n.

        A sibling node is defined as a node that:
        1. shares the same parent of `n`
        2. shares at least one input of `n`

        Example:

              P
              |  (e0)
             / \
            A   B

            Say node A is the sibling of node B, and vice versa. The common edge
            is [e0].
        """
        assert isinstance(n, (onnx.NodeProto, str))
        parents_and_edges: Dict[onnx.NodeProto, str] = {}
        if isinstance(n, str):
            node_name = n
        else:
            node_name = n.name
        for parent in self.onnx_predecessors(n):
            edge_data = self.get_edge_data(parent.name, node_name)
            parents_and_edges[parent] = str(edge_data["edge"])

        def _child_and_edge(node: onnx.NodeProto, edge: str):
            for child in self.onnx_successors(node):
                if edge == self.get_edge_data(node.name, child.name).get("edge"):
                    yield child

        siblings: List[onnx.NodeProto] = []
        for parent, edge in parents_and_edges.items():
            for child in _child_and_edge(parent, edge):
                if child.name != node_name:
                    siblings.append(child)
        return siblings

    def onnx_subgraph(self, nodes: Iterable[onnx.NodeProto | str]) -> "OnnxGraph":
        """Create a sub onnx graph from nodes"""
        sub: nx.DiGraph = self.subgraph(  # type: ignore
            (n if isinstance(n, str) else n.name for n in nodes)
        )
        subonnx = make_graph([], f"subgraph of {self.name}", inputs=[], outputs=[])
        # inherite initializers and value infos
        initializer_map = {i.name: i for i in self.initializer}
        subonnx.value_info.extend(self._value_info)

        def _compute_input(node):
            check_list = set(initializer_map.keys())
            for pred in sub.predecessors(node.name):
                check_list.update(sub.nodes[pred]["pb"].output)
            for input_name in node.input:
                if input_name and input_name not in check_list:
                    yield input_name
                if input_name in initializer_map:
                    subonnx.initializer.append(initializer_map[input_name])
                    initializer_map.pop(input_name)  # avoid duplicated initializer

        graph_inputs = set()
        for i in nx.topological_sort(sub):
            node = self.nodes[i]["pb"]
            subonnx.node.append(node)
            for output_name in node.output:
                is_outside = any(  # has an edge connected outside subgraph
                    output_name in j.input
                    for j in self.onnx_successors(node)
                    if j.name not in sub
                )
                # PR#176: if node is the output of the graph, we should not cut
                # it and must keep this output in the subgraph
                is_outside |= output_name in self.outputs
                if sub.out_degree(i) == 0 or is_outside:
                    shape, dtype = self.tensor_info(output_name)
                    subonnx.output.append(
                        make_value_info(
                            output_name, make_tensor_type_proto(dtype, shape)
                        )
                    )
            node_inputs = list(_compute_input(node))
            if node.op_type != "Constant" and node_inputs:
                graph_inputs.update(node_inputs)
        sorted_graph_inputs = sorted(graph_inputs)
        for input_name in sorted_graph_inputs:
            shape, dtype = self.tensor_info(input_name)
            subonnx.input.append(
                make_value_info(input_name, make_tensor_type_proto(dtype, shape))
            )
        sub_model = make_model(
            subonnx,
            doc_string=self._model.doc_string,
            domain=self._model.domain,
            model_version=self._model.model_version,
            ir_version=self._model.ir_version,
            opset_imports=self._model.opset_import,
            producer_name=self._model.producer_name,
            producer_version=self._model.producer_version,
            functions=self.functions.values(),
        )
        return OnnxGraph(sub_model, base_dir=self.external_base)

    def onnx_add_function(self, func: onnx.FunctionProto, force: bool = False) -> None:
        """Add a function to the graph.

        Args:
            func (onnx.FunctionProto): an onnx function proto.
            force (bool, optional): whether to forcely add the function.
                Defaults to False.
        """
        if func.name not in self._functions or force:
            self._functions[func.name] = func
            self._model.opset_import.extend(func.opset_import)

    def tensor_info(self, name: str) -> Tuple[List[int | str] | None, int]:
        """Get shape and dtype of a tensor by its name."""
        shape: Optional[List[int | str]] = None
        dtype = onnx.TensorProto.UNDEFINED
        if name in self.inputs or name in self.outputs:
            value_infos = list(chain(self.input, self.output))
        else:
            value_infos = self._value_info
        for value_info in value_infos:
            if value_info.name == name:
                if value_info.type.HasField("tensor_type"):
                    tensor_type = value_info.type.tensor_type
                elif value_info.type.HasField("sequence_type"):
                    seq_type = value_info.type.sequence_type.elem_type
                    # for now, sequence type must be a tensor type
                    if not seq_type.HasField("tensor_type"):
                        raise TypeError(f"Unsupported sequence type: {seq_type}")
                    tensor_type = seq_type.tensor_type
                else:
                    raise TypeError(f"Unsupported value type: {value_info.type}")
                dtype = tensor_type.elem_type
                if not tensor_type.HasField("shape"):
                    break  # no shape info, skip
                dims = [i for i in tensor_type.shape.dim]
                shape = [0 for _ in dims]
                for i, s in enumerate(dims):
                    if s.dim_param and s.dim_value == 0:
                        shape[i] = s.dim_param
                    else:
                        shape[i] = int(s.dim_value)
                break
        if shape is None:
            for tensor in self.initializer:
                if tensor.name == name:
                    shape = [int(i) for i in tensor.dims]
                    dtype = tensor.data_type
        return shape, dtype

    def tensor_shape(self, name: str) -> List[int | str]:
        """Get shape from a given tensor name."""
        shape, _ = self.tensor_info(name)
        if shape is None:
            raise ValueError(f"Can't find tensor shape with name '{name}'")
        return shape

    def static_tensor_shape(self, name: str) -> List[int]:
        """Get shape from a given tensor name, and ensure it is static."""
        shape = self.tensor_shape(name)
        static_shape: List[int] = []
        for ind, i in enumerate(shape):
            if isinstance(i, str) or i < 1:
                raise ValueError(f"shape[{ind}] is dynamic: {i}")
            static_shape.append(i)
        return static_shape

    def tensor_type(self, name: str) -> int:
        """Get tensor type from a given tensor name."""
        _, elem_type = self.tensor_info(name)
        if elem_type == onnx.TensorProto.UNDEFINED:
            raise ValueError(f"Can't find tensor dtype with name '{name}'")
        return elem_type

    def set_input(self, node: onnx.NodeProto | str, input_name: str):
        """Set the `input_name` of `node` as the graph input."""
        if isinstance(node, str):
            node = self.nodes[node]["pb"]
        assert isinstance(node, onnx.NodeProto)
        if input_name not in node.input:
            raise ValueError(f"Can't find {input_name} in {node.name}")
        if input_name in self.inputs:
            return  # already input
        shape, dtype = self.tensor_info(input_name)
        self.inputs[input_name] = len(self.input)
        if shape is None or dtype is None:
            raise ValueError(f"Can't find tensor shape and dtype of {input_name}")
        self.input.append(
            make_value_info(input_name, make_tensor_type_proto(dtype, shape))
        )

    def set_output(self, node: onnx.NodeProto | str, output_name: str):
        """Set the `output_name` of `node` as the graph output."""
        if isinstance(node, str):
            node = self.nodes[node]["pb"]
        assert isinstance(node, onnx.NodeProto)
        if output_name not in node.output:
            raise ValueError(f"Can't find {output_name} in {node.name}")
        replace = output_name in self.outputs
        shape, dtype = self.tensor_info(output_name)
        if replace:
            self.remove_output(output_name)
        self.outputs[output_name] = len(self.output)
        self._out_to_node[output_name] = node.name
        if shape is None or dtype is None:
            raise ValueError(f"Can't find tensor shape and dtype of {output_name}")
        self.output.append(
            make_value_info(output_name, make_tensor_type_proto(dtype, shape))
        )

    def remove_input(self, input_name: str):
        """Remove the `input_name` from the graph input."""
        ind = self.inputs.pop(input_name)
        self.input.pop(ind)
        for i in self.inputs:
            if self.inputs[i] > ind:
                self.inputs[i] -= 1

    def remove_output(self, output_name: str):
        """Remove the `output_name` from the graph output."""
        ind = self.outputs.pop(output_name)
        self.output.pop(ind)
        for i in self.outputs:
            if self.outputs[i] > ind:
                self.outputs[i] -= 1

    def rename_input(self, old_name: str, new_name: str):
        """Rename the `old_name` input to `new_name`."""
        self.inputs[new_name] = self.inputs.pop(old_name)
        self.input[self.inputs[new_name]].name = new_name
        for n in nx.topological_sort(self):
            if self.in_degree(n) > 0:  # type: ignore
                break
            input_node = self.nodes[n]["pb"]
            for ind, i in enumerate(input_node.input):
                if i == old_name:
                    input_node.input[ind] = new_name

    def rename_output(self, old_name: str, new_name: str):
        """Rename the `old_name` output to `new_name`."""
        self.outputs[new_name] = self.outputs.pop(old_name)
        self.output[self.outputs[new_name]].name = new_name
        self._out_to_node[new_name] = self._out_to_node.pop(old_name)
        output_node_name = self._out_to_node[new_name]
        output_node = self.nodes[output_node_name]["pb"]
        for ind, i in enumerate(output_node.output):
            if i == old_name:
                output_node.output[ind] = new_name
        self._node_to_out[output_node.name] = output_node.output

    def set_value_info(
        self, name: str, shape: Sequence[int | str], dtype: Optional[int] = None
    ):
        """Overwrite the value info of a tensor.

        Args:
            name (str): the name of the tensor.
            shape (List[int | str]): the shape of the tensor.
            dtype (int, optional): the data type of the tensor. Defaults to None.
        """
        for i, info in enumerate(self._value_info):
            if info.name == name:
                if dtype is None:
                    dtype = info.type.tensor_type.elem_type
                self._value_info.pop(i)
                self._value_info_update.append(
                    make_tensor_value_info(name, dtype, shape)
                )
                return
        if dtype is None:
            raise ValueError(
                f"dtype is required because value is not existing for {name}"
            )
        self._value_info_update.append(make_tensor_value_info(name, dtype, shape))

    def set_seqeuence_info(
        self, name: str, shape: Sequence[int | str], dtype: Optional[int] = None
    ):
        """Overwrite the value info of a sequence.

        Args:
            name (str): the name of the sequence.
            shape (List[int | str]): the shape of the tensor in the sequence.
            dtype (int, optional): the data type of the tensor. Defaults to None.
        """
        for i, info in enumerate(self._value_info):
            if info.name == name:
                if dtype is None:
                    dtype = info.type.tensor_type.elem_type
                self._value_info.pop(i)
                self._value_info_update.append(
                    make_tensor_sequence_value_info(name, dtype, shape)
                )
                return
        if dtype is None:
            raise ValueError(
                f"dtype is required because value is not existing for {name}"
            )
        self._value_info_update.append(
            make_tensor_sequence_value_info(name, dtype, shape)
        )

    @property
    def input(self):
        """Return a list of graph inputs value info."""
        return self._model.graph.input

    @property
    def output(self):
        """Return a list of graph outputs value info."""
        return self._model.graph.output

    @property
    def initializer(self):
        """Return a list of graph initializer tensors."""
        return self._model.graph.initializer

    @property
    def initializers(self) -> Dict[str, onnx.TensorProto]:
        """Return a dict of graph initializer tensors."""
        return {i.name: i for i in self.initializer}

    @property
    def model_version(self) -> int:
        """Return the model version of the model."""
        return self._model.model_version

    @property
    def ir_version(self) -> int:
        """Return the ir version of the model."""
        return self._model.ir_version

    @property
    def opset_version(self) -> int:
        """Return the opset version of the model."""
        return max(
            i.version for i in self._model.opset_import if i.domain in ("", "ai.onnx")
        )

    @opset_version.setter
    def opset_version(self, version: int):
        """Set the opset version of the model."""
        # keep only one ai.onnx opset
        for opset in list(self._model.opset_import):
            if opset.domain in ("", "ai.onnx"):
                self._model.opset_import.remove(opset)
        self._model.opset_import.append(make_operatorsetid("", version))

    @property
    def functions(self) -> Dict[str, onnx.FunctionProto]:
        """Return the functions inside the model."""
        return self._functions

    @property
    def external_base(self) -> Optional[str]:
        """Return the base directory of external data."""
        return self._external_base

    @external_base.setter
    def external_base(self, base_dir: str | os.PathLike | None):
        """Set the base directory of external data."""
        if base_dir:
            self._external_base = self._get_external_base(str(base_dir))
        else:
            self._external_base = self._get_external_base(None)

    @property
    def external_data(self) -> tuple[Path, ...]:
        """Yield all external data file locations."""

        if not self._external_base:
            return ()
        data = set()
        for t in self._get_all_tensors():
            if edh.uses_external_data(t):
                info = edh.ExternalDataInfo(t)
                location = Path(self._external_base) / info.location
                data.add(location)
        return tuple(data)

    def _get_all_tensors(self) -> Iterable[onnx.TensorProto]:
        # pylint: disable=protected-access
        yield from self.initializer
        for func in self.functions.values():
            yield from edh._get_attribute_tensors_from_graph(func)
        for name in self:
            node: onnx.NodeProto = self.nodes[name]["pb"]
            for attribute in node.attribute:
                if attribute.HasField("t"):
                    yield attribute.t
                yield from attribute.tensors
                yield from edh._recursive_attribute_processor(
                    attribute, edh._get_initializer_tensors_from_graph
                )
                yield from edh._recursive_attribute_processor(
                    attribute, edh._get_attribute_tensors_from_graph
                )

    def convert_tensors_to_external(
        self,
        location: str | os.PathLike | Path,
        size_threshold: int = 1024,
    ):
        """Set all tensors with raw data as external data.

        Args:
            location (str | os.PathLike | Path): specify the external file to save to.
            size_threshold (int): Threshold for size of data. Only when
                tensor's data is >= the size_threshold it will be converted to external
                data. To convert every tensor with raw data to external data set
                size_threshold=0. Defaults to 1024.
        """
        location = Path(location)
        if not self._external_base:
            self._external_base = Path.cwd().as_posix()
            if location.is_absolute():
                self._external_base = location.parent.as_posix()
        if location.is_absolute() and location.is_relative_to(self._external_base):
            location = location.relative_to(self._external_base)
        elif location.is_absolute():
            raise ValueError(f"absolute location {location} must be relative to cwd.")
        # recreate the file
        with open(self._external_base / location, "wb"):
            pass
        for tensor in self._get_all_tensors():
            if tensor.HasField("raw_data") and len(tensor.raw_data) >= size_threshold:
                edh.set_external_data(tensor, location.as_posix())
        for tensor in self._get_all_tensors():
            if edh.uses_external_data(tensor) and tensor.HasField("raw_data"):
                edh.save_external_data(tensor, self._external_base)
                tensor.ClearField("raw_data")
        # free memory
        gc.collect()

    def restore_tensors_from_external(self):
        """Restore all external tensors into graph."""
        if not self._external_base:
            return

        for tensor in self._get_all_tensors():
            if edh.uses_external_data(tensor):
                info = edh.ExternalDataInfo(tensor)
                location = Path(self._external_base) / info.location
                with open(location, "rb") as data_file:
                    if info.offset:
                        data_file.seek(info.offset)

                    if info.length:
                        tensor.raw_data = data_file.read(info.length)
                    else:
                        tensor.raw_data = data_file.read()
                # After loading raw_data from external_data, change the state of tensors
                tensor.data_location = onnx.TensorProto.DEFAULT
                # and remove external data
                del tensor.external_data[:]
        self._external_base = None

    def canonicalize_external_tensors(self):
        """When `numpy_helper.to_array` is called, a tensor with external data
        is loaded into .raw_data filed. We need to remove this filed to pass
        the checker and also avoid saving extra data into onnx model.
        """
        for tensor in self._get_all_tensors():
            if edh.uses_external_data(tensor) and tensor.HasField("raw_data"):
                tensor.ClearField("raw_data")

    @property
    def model(self) -> onnx.ModelProto:
        """Make new model"""
        graph = deepcopy(self._model.graph)
        graph.ClearField("node")
        if self._keep_value_info:
            # remove duplicated values, later value_info will overwrite earlier ones
            values_map = {
                i.name: i for i in chain(self._value_info, self._value_info_update)
            }
            self._value_info_update = list(values_map.values())
        graph.ClearField("value_info")
        for n in nx.topological_sort(self):
            graph.node.append(self.nodes[n]["pb"])
        graph.value_info.extend(self._value_info_update)
        model = make_model(
            graph,
            doc_string=self._model.doc_string,
            domain=self._model.domain,
            model_version=self._model.model_version,
            ir_version=self._model.ir_version,
            opset_imports=_unique_opset(self._model.opset_import),
            producer_name=self._model.producer_name,
            producer_version=self._model.producer_version,
            functions=self._functions.values(),
        )
        model.metadata_props.extend(self._model.metadata_props)
        return model

    # pylint: disable=redefined-builtin
    def save(
        self,
        model_path: str | os.PathLike | IO[bytes],
        format: Optional[str] = None,
        save_as_external_data: bool = False,
        infer_shapes: bool = True,
        check: bool = True,
    ):
        """Serialize the graph to onnx model and save to model_path."""
        if save_as_external_data:
            if isinstance(model_path, (IO, IOBase)):
                location = (
                    self.name.replace(":", "_")
                    .replace("/", "_")
                    .replace("?", "")
                    .replace("!", "")
                    .replace(" ", "")
                )
            else:
                location = Path(model_path).resolve().with_suffix("")
            self.restore_tensors_from_external()
            self.convert_tensors_to_external(location=location)
        else:
            self.canonicalize_external_tensors()
        if infer_shapes:
            # infer shape using infer_shape pass
            # pylint: disable=import-outside-toplevel
            from .passes.globals.infer_shape import infer_shape

            graph_with_shape = infer_shape(self)
            # pylint: disable=protected-access
            graph_with_shape._keep_value_info = True
            model = graph_with_shape.model
        else:
            model = self.model
        if not isinstance(model_path, (str, os.PathLike)):
            format = "protobuf"
        elif format == "protobuf" or format is None:
            model_path = Path(model_path).with_suffix(".onnx")
        elif format == "textproto":
            model_path = Path(model_path).with_suffix(".pbtxt")
        elif format == "json":
            model_path = Path(model_path).with_suffix(".json")
        elif format == "onnxtxt":
            model_path = Path(model_path).with_suffix(".onnxtxt")
        if check:
            with chdir(self.external_base):
                try:
                    onnx.checker.check_model(model, full_check=True)
                except onnx.checker.ValidationError as ex:
                    error(f"onnx check failed (use -vv to show tracebacks):\n{ex}\n")
                    debug("\n".join(traceback.format_exception(ex)))
        onnx.save_model(model, model_path, format=format)
        return model_path



"""
Copyright (C) 2025 The ONNXIFIER Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from typing import List, cast

import numpy as np
from onnx import NodeProto
from onnx.helper import make_node

from ...graph import OnnxGraph
from .. import PASSES
from ..pattern import SingleNodePattern
from ..rewriter import Rewriter
from ..utils import make_constant


@PASSES.register("avgpool_to_conv")
class AvgPoolToConvRewriter(Rewriter):
    """Replace eligible AveragePool with group-wise Conv.

    Conditions for conversion:
    - Operator is `AveragePool` (not `GlobalAveragePool`).
    - `ceil_mode` is 0 or not set.
    - If `count_include_pad` is 0, then all pads must be 0.
      If `count_include_pad` is 1, arbitrary pads are supported.

    Result:
    - Replace with `Conv` with group = C and weights filled with 1/(prod(kernel))
      of shape (C, 1, *kernel_shape). Strides/pads/auto_pad are preserved.
    """

    def __init__(self):
        super().__init__(SingleNodePattern("AveragePool"))

    def rewrite(self, graph: OnnxGraph, nodes: List[NodeProto]):
        node = nodes[0]

        # Extract attributes
        kernel_shape_attr = self.get_attribute(node, "kernel_shape")
        if (
            not isinstance(kernel_shape_attr, list)
            or len(kernel_shape_attr) == 0
            or not all(isinstance(k, (int, np.integer)) for k in kernel_shape_attr)
        ):
            return
        kernel_shape = [int(k) for k in kernel_shape_attr]  # type: ignore
        spatial_rank = len(kernel_shape)

        strides_attr = self.get_attribute(node, "strides", None)
        if isinstance(strides_attr, list) and len(strides_attr) == spatial_rank:
            validated_strides: List[int] = []
            for s in strides_attr:
                if isinstance(s, (int, np.integer)):
                    validated_strides.append(int(cast(int, s)))
                else:
                    validated_strides = []
                    break
            if validated_strides:
                strides = validated_strides
            else:
                strides = [1] * spatial_rank
        else:
            strides = [1] * spatial_rank

        pads_attr = self.get_attribute(node, "pads", None)
        if isinstance(pads_attr, list) and len(pads_attr) == 2 * spatial_rank:
            validated_pads: List[int] = []
            for p in pads_attr:
                if isinstance(p, (int, np.integer)):
                    validated_pads.append(int(cast(int, p)))
                else:
                    validated_pads = []
                    break
            if validated_pads:
                pads = validated_pads
            else:
                pads = [0] * (2 * spatial_rank)
        else:
            pads = [0] * (2 * spatial_rank)

        auto_pad_attr = self.get_attribute(node, "auto_pad", "NOTSET")
        auto_pad = str(auto_pad_attr) if auto_pad_attr is not None else "NOTSET"

        ceil_mode_attr = self.get_attribute(node, "ceil_mode", 0)
        ceil_mode = (
            int(ceil_mode_attr) if isinstance(ceil_mode_attr, (int, np.integer)) else 0
        )

        cip_attr = self.get_attribute(node, "count_include_pad", 0)
        count_include_pad = (
            int(cip_attr) if isinstance(cip_attr, (int, np.integer)) else 0
        )

        # Unsupported settings
        if ceil_mode != 0:
            return
        if count_include_pad == 0 and any(int(p) != 0 for p in pads):
            # Excluding padded zeros cannot be modeled with a single fixed Conv
            return

        # Validate input shape and channels
        input_shape = graph.tensor_shape(node.input[0])
        if input_shape is None or len(input_shape) < spatial_rank + 2:
            return
        channels_any = input_shape[1]
        if not isinstance(channels_any, int) or channels_any <= 0:
            return
        channels: int = int(channels_any)

        # Build depthwise weights: (C, 1, *kernel) with average factor
        kernel_prod = int(np.prod([int(k) for k in kernel_shape]))
        scale = 1.0 / float(kernel_prod)
        weight_value = np.full(
            (channels, 1, *[int(k) for k in kernel_shape]), scale, dtype=np.float32
        )
        weight_cst = make_constant(f"{node.name}/weights", weight_value)

        conv_kwargs: dict = dict(strides=[int(s) for s in strides], group=channels)
        if auto_pad != "NOTSET":
            conv_kwargs["auto_pad"] = auto_pad
        else:
            conv_kwargs["pads"] = [int(p) for p in pads]

        conv = make_node(
            "Conv",
            inputs=[node.input[0], weight_cst.output[0]],
            outputs=list(node.output),
            name=node.name,
            **conv_kwargs,
        )

        self += [weight_cst, conv]
        self -= node
